# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import configparser
import enum
import io
import logging
import os
import sys
import urllib.parse
from typing import TextIO

from xdg import BaseDirectory

from snapcraft.internal.errors import SnapcraftInvalidCLIConfigError
from snapcraft.storeapi import constants, errors

LOCAL_CONFIG_FILENAME = ".snapcraft/snapcraft.cfg"

logger = logging.getLogger(__name__)


@enum.unique
class OutdatedStepAction(enum.Enum):
    # Would like to use enum.auto(), but it's only available in >= 3.6
    ERROR = 1
    CLEAN = 2


class CLIConfig:
    """Hold general options affecting the CLI.

    Setting data takes the convention of set_<section>_<option> and
    getting data takes a similar form get_<section>_<option>.
    """

    def __init__(self, *, read_only: bool = False) -> None:
        """Initialize an instance of CLIConfig.

        :param bool read_only: if set, attempts to set data will fail.
        """
        self.parser = configparser.ConfigParser()
        self.config_path = os.path.join(
            BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
        )
        self._read_only = read_only

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._read_only:
            self.save()

    def load(self) -> None:
        """Load config parser data.

        :raises SnapcraftInvalidCLIConfigError:
            if the configuration data cannot be loaded from disk.
        """
        try:
            with open(self.config_path) as f:
                config = f.read()
        except FileNotFoundError:
            config = None

        if config:
            try:
                self.parser.read_string(config)
            except configparser.Error as parse_error:
                raise SnapcraftInvalidCLIConfigError(
                    config_file=self.config_path, error=str(parse_error)
                ) from parse_error

    def save(self) -> None:
        """Save config parser data.

        :raises RuntimeError: if trying to save when operating with read_only.
        """
        if self._read_only:
            raise RuntimeError("CLIConfig was instantiated with read_only.")
        # Make sure the directory exists before saving
        os.makedirs(os.path.dirname(self.config_path), exist_ok=True)

        with open(self.config_path, "w") as f:
            self.parser.write(f)

    def _get_option(self, section_name: str, option_name: str) -> str:
        try:
            return self.parser.get(section_name, option_name)
        except (configparser.NoSectionError, configparser.NoOptionError, KeyError):
            return None

    def _set_option(self, section_name, option_name: str, value: str) -> None:
        if self._read_only:
            raise RuntimeError("CLIConfig was instantiated with read_only.")

        section_name = section_name
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)
        self.parser.set(section_name, option_name, value)

    def set_sentry_send_always(self, value: bool) -> None:
        """Setter to define if errors should automatically be sent to sentry.

        :param bool value: if true, errors should automatically be sent.
        """
        if value:
            string_value = "true"
        else:
            string_value = "false"
        self._set_option("Sentry", "always_send", string_value)

    def get_sentry_send_always(self) -> bool:
        """Getter to define if errors should automatically be sent to sentry.

        :returns: True if errors should automatically be sent.
        :rtype: bool.
        """
        string_value = self._get_option("Sentry", "always_send")
        # Anything but "true" for string_value is considered False.
        return string_value == "true"

    def set_outdated_step_action(self, action: OutdatedStepAction) -> None:
        """Setter to define action to take if outdated step is encountered.

        :param OutdatedStepAction value: The action to take
        """
        self._set_option("Lifecycle", "outdated_step_action", action.name.lower())

    def get_outdated_step_action(self) -> OutdatedStepAction:
        """Getter to define action to take if outdated step is encountered.

        :returns: The action to take
        :rtype: OutdatedStepAction.
        """
        action = self._get_option("Lifecycle", "outdated_step_action")
        if action:
            return OutdatedStepAction[action.upper()]
        else:
            # Clean by default
            return OutdatedStepAction.CLEAN


class Config(object):
    """Hold configuration options in sections.

    There can be two sections for the sso related credentials: production and
    staging. This is governed by the UBUNTU_SSO_API_ROOT_URL environment
    variable. Other sections are ignored but preserved.

    """

    def __init__(self) -> None:
        self.parser = configparser.ConfigParser()
        self.load()

    def _section_name(self) -> str:
        # The only section we care about is the host from the SSO url
        url = os.environ.get(
            "UBUNTU_SSO_API_ROOT_URL", constants.UBUNTU_SSO_API_ROOT_URL
        )
        return urllib.parse.urlparse(url).netloc

    def get(self, option_name: str) -> str:
        try:
            return self.parser.get(self._section_name(), option_name)
        except (configparser.NoSectionError, configparser.NoOptionError, KeyError):
            return None

    def set(self, option_name: str, value: str) -> None:
        section_name = self._section_name()
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)
        self.parser.set(section_name, option_name, value)

    def is_empty(self) -> bool:
        # Only check the current section
        section_name = self._section_name()
        if self.parser.has_section(section_name):
            if self.parser.options(section_name):
                return False
        return True

    def load(self, *, config_fd: TextIO = None) -> None:
        config = ""
        if config_fd:
            config = config_fd.read()
        else:
            # Local configurations (per project) are supposed to be static.
            # That's why it's only checked for 'loading' and never written to.
            # Essentially, all authentication-related changes, like
            # login/logout or macaroon-refresh, will not be persisted for the
            # next runs.
            file_path = ""
            if os.path.exists(LOCAL_CONFIG_FILENAME):
                file_path = LOCAL_CONFIG_FILENAME

                # FIXME: We don't know this for sure when loading the config.
                # Need a better separation of concerns.
                logger.warn(
                    "Using local configuration ({!r}), changes will not be "
                    "persisted.".format(file_path)
                )
            else:
                file_path = BaseDirectory.load_first_config(
                    "snapcraft", "snapcraft.cfg"
                )
            if file_path and os.path.exists(file_path):
                with open(file_path, "r") as f:
                    config = f.read()

        if config:
            _load_potentially_base64_config(self.parser, config)

    @staticmethod
    def save_path() -> str:
        return os.path.join(
            BaseDirectory.save_config_path("snapcraft"), "snapcraft.cfg"
        )

    def save(self, *, config_fd: TextIO = None, encode: bool = False) -> None:
        with io.StringIO() as config_buffer:
            self.parser.write(config_buffer)
            config = config_buffer.getvalue()
            if encode:
                # Encode config using base64
                config = base64.b64encode(
                    config.encode(sys.getfilesystemencoding())
                ).decode(sys.getfilesystemencoding())

            if config_fd:
                config_fd.write(config)
            else:
                with open(self.save_path(), "w") as f:
                    f.write(config)

    def clear(self) -> None:
        self.parser.remove_section(self._section_name())


def _load_potentially_base64_config(parser, config):
    try:
        parser.read_string(config)
    except configparser.Error as e:
        # The config may be base64-encoded, try decoding it
        try:
            config = base64.b64decode(config).decode(sys.getfilesystemencoding())
        except base64.binascii.Error:  # type: ignore
            # It wasn't base64, so use the original error
            raise errors.InvalidLoginConfig(e) from e

        try:
            parser.read_string(config)
        except configparser.Error as e:
            raise errors.InvalidLoginConfig(e) from e



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import fileinput
import glob
import itertools
import logging
import os
import re
import shutil
import stat
from typing import List

from snapcraft import file_utils
from snapcraft.internal import mangling
from . import errors


logger = logging.getLogger(__name__)


class BaseRepo:
    """Base implementation for a platform specific repo handler.

    Generally any new repo handling system would inherit from this and
    implement:

    - get
    - unpack
    - get_package_libraries
    - get_packages_for_source_type
    - refresh_build_packages
    - install_build_packages
    - is_package_installed

    At the end of the `unpack` method `normalize` needs to be called to
    adapt the artifacts downloaded to be generic enough for building a snap."""

    @classmethod
    def get_package_libraries(cls, package_name):
        """Return a list of libraries in package_name.

        Given the contents of package_name, return the subset of what are
        considered libraries from those contents, be it static or shared.

        :param str package: package name to get library contents from.
        :returns: a list of libraries that package_name provides. This includes
                  directories.
        :rtype: set.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_package_for_file(cls, file_path: str) -> str:
        """Return the package name that provides file_path.

        :param str file_path: the absolute path to the file to search for.
        :returns: package name that provides file_path.
        :rtype: str
        :raises snapcraft.repo.errors.FileProviderNotFound:
            if file_path is not provided by any package.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_packages_for_source_type(cls, source_type):
        """Return a list of packages required to to work with source_type.

        source_type can be any of the following:

        - bzr
        - deb
        - rpm
        - git
        - hg
        - mercurial
        - subversion
        - svn
        - tar
        - zip

        :param str source_type: a VCS source type to handle.
        :returns: a set of packages that need to be installed on the host.
        :rtype: set of strings.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def refresh_build_packages(cls) -> None:
        """Refresh the build packages cache.

        If refreshing is not possible
        snapcraft.repo.errors.CacheUpdateFailedError should be raised

        :raises snapcraft.repo.errors.NoNativeBackendError:
            if the method is not implemented in the subclass.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def install_build_packages(cls, package_names: List[str]) -> List[str]:
        """Install packages on the host required to build.

        This method needs to be implemented by using the appropriate method
        to install packages on the system. If possible they should be marked
        as automatically installed to allow for easy removal.
        The method should return a list of the actually installed packages
        in the form "package=version".

        If one of the packages cannot be found
        snapcraft.repo.errors.BuildPackageNotFoundError should be raised.
        If dependencies for a package cannot be resolved
        snapcraft.repo.errors.PackageBrokenError should be raised.
        If installing a package on the host failed
        snapcraft.repo.errors.BuildPackagesNotInstalledError should be raised.

        :param package_names: a list of package names to install.
        :type package_names: a list of strings.
        :return: a list with the packages installed and their versions.
        :rtype: list of strings.
        :raises snapcraft.repo.errors.NoNativeBackendError:
            if the method is not implemented in the subclass.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def build_package_is_valid(cls, package_name):
        """Check that a given package is valid on the host.

        :param package_name: a package name to check.
        :type package_name: str
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def is_package_installed(cls, package_name):
        """Return a bool indicating if package_name is installed.

        :param str package_name: the package name to query.
        :returns: True if package_name is installed if not False.
        :rtype: boolean
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_installed_packages(cls):
        """Return a list of the installed packages and their versions

        :rtype: list of strings with the form package=version.
        """
        raise errors.NoNativeBackendError()

    def __init__(self, rootdir, *args, **kwargs):
        """Initialize a repository handler.

        :param str rootdir: the root directory to work on. This may well be to
                            create repo handling specific caches or download
                            artifacts.
        """
        self.rootdir = rootdir

    def get(self, package_names):
        """Get the packages.

        This method needs to be implemented by the inheriting class. It
        is called all along the core of snapcraft to obtain the
        package_names and store them in some repo defined location to
        later unpack.

        :param str package_names: list of packages to obtain.
        :returns: list of packages obtained, versioned if possible.
        :rtype: list of strings.
        :raises snapcraft.repo.errors.PackageNotFoundError:
            when a package in package_names is not found.
        """
        raise errors.NoNativeBackendError()

    def unpack(self, unpackdir):
        """Unpack obtained packages into unpackdir.

        This method needs to be implemented by the inheriting class. It
        is called all along the core of snapcraft to unpack the previously
        obtained package_names in get into unpackdir.

        After the unpack logic is executed, normalize should be called.

        :param str unpackdir: target directory to unpack packages to.
        """
        raise errors.NoNativeBackendError()

    def normalize(self, unpackdir):
        """Normalize artifacts in unpackdir.

        Repo specific packages are generally created to live in a specific
        distro. What normalize does is scan through the unpacked artifacts
        and slightly modifies them to work better with snapcraft projects
        when building and to also work within a snap's environment.

        :param str unpackdir: directory where files where unpacked.
        """
        self._remove_useless_files(unpackdir)
        self._fix_artifacts(unpackdir)
        self._fix_xml_tools(unpackdir)
        self._fix_shebangs(unpackdir)

    def _remove_useless_files(self, unpackdir):
        """Remove files that aren't useful or will clash with other parts."""
        sitecustomize_files = glob.glob(
            os.path.join(unpackdir, "usr", "lib", "python*", "sitecustomize.py")
        )
        for sitecustomize_file in sitecustomize_files:
            os.remove(sitecustomize_file)

    def _fix_artifacts(self, unpackdir):
        """Perform various modifications to unpacked artifacts.

        Sometimes distro packages will contain absolute symlinks (e.g. if the
        relative path would go all the way to root, they just do absolute). We
        can't have that, so instead clean those absolute symlinks.

        Some unpacked items will also contain suid binaries which we do not
        want in the resulting snap.
        """
        for root, dirs, files in os.walk(unpackdir):
            # Symlinks to directories will be in dirs, while symlinks to
            # non-directories will be in files.
            for entry in itertools.chain(files, dirs):
                path = os.path.join(root, entry)
                if os.path.islink(path) and os.path.isabs(os.readlink(path)):
                    self._fix_symlink(path, unpackdir, root)
                elif os.path.exists(path):
                    _fix_filemode(path)

                if path.endswith(".pc") and not os.path.islink(path):
                    fix_pkg_config(unpackdir, path)

    def _fix_xml_tools(self, unpackdir):
        xml2_config_path = os.path.join(unpackdir, "usr", "bin", "xml2-config")
        with contextlib.suppress(FileNotFoundError):
            file_utils.search_and_replace_contents(
                xml2_config_path,
                re.compile(r"prefix=/usr"),
                "prefix={}/usr".format(unpackdir),
            )

        xslt_config_path = os.path.join(unpackdir, "usr", "bin", "xslt-config")
        with contextlib.suppress(FileNotFoundError):
            file_utils.search_and_replace_contents(
                xslt_config_path,
                re.compile(r"prefix=/usr"),
                "prefix={}/usr".format(unpackdir),
            )

    def _fix_symlink(self, path, unpackdir, root):
        host_target = os.readlink(path)
        if host_target in self.get_package_libraries("libc6"):
            logger.debug(
                "Not fixing symlink {!r}: it's pointing to libc".format(host_target)
            )
            return

        target = os.path.join(unpackdir, os.readlink(path)[1:])
        if not os.path.exists(target) and not _try_copy_local(path, target):
            return
        os.remove(path)
        os.symlink(os.path.relpath(target, root), path)

    def _fix_shebangs(self, unpackdir):
        """Change hard-coded shebangs in unpacked files to use env."""
        mangling.rewrite_python_shebangs(unpackdir)


class DummyRepo(BaseRepo):
    def get_packages_for_source_type(*args, **kwargs):
        return set()


def _try_copy_local(path, target):
    real_path = os.path.realpath(path)
    if os.path.exists(real_path):
        logger.warning(
            "Copying needed target link from the system {}".format(real_path)
        )
        os.makedirs(os.path.dirname(target), exist_ok=True)
        shutil.copyfile(os.readlink(path), target)
        return True
    else:
        logger.warning("{} will be a dangling symlink".format(path))
        return False


def fix_pkg_config(root, pkg_config_file, prefix_trim=None):
    """Opens a pkg_config_file and prefixes the prefix with root."""
    pattern_trim = None
    if prefix_trim:
        pattern_trim = re.compile("^prefix={}(?P<prefix>.*)".format(prefix_trim))
    pattern = re.compile("^prefix=(?P<prefix>.*)")

    with fileinput.input(pkg_config_file, inplace=True) as input_file:
        for line in input_file:
            match = pattern.search(line)
            if prefix_trim:
                match_trim = pattern_trim.search(line)
            if prefix_trim and match_trim:
                print("prefix={}{}".format(root, match_trim.group("prefix")))
            elif match:
                print("prefix={}{}".format(root, match.group("prefix")))
            else:
                print(line, end="")


def _fix_filemode(path):
    mode = stat.S_IMODE(os.stat(path, follow_symlinks=False).st_mode)
    if mode & 0o4000 or mode & 0o2000:
        logger.warning("Removing suid/guid from {}".format(path))
        os.chmod(path, mode & 0o1777)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
from typing import Dict, Optional, Sequence

from .. import errors
from .._base_provider import Provider
from ._instance_info import InstanceInfo
from ._multipass_command import MultipassCommand
from snapcraft.internal.errors import SnapcraftEnvironmentError


logger = logging.getLogger(__name__)


def _get_stop_time() -> int:
    try:
        timeout = int(os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME", "10"))
    except ValueError:
        raise SnapcraftEnvironmentError(
            "'SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME' is incorrectly set, found {!r} "
            "but expected a number representing the minutes to delay the actual stop "
            "operation (or 0 to stop immediately).".format(
                os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME")
            )
        )

    return timeout


class _MachineSetting:
    def __init__(self, *, envvar: str, default: str) -> None:
        self._default = default
        self._envvar = envvar

    def get_value(self):
        value = os.getenv(self._envvar, self._default)
        if value != self._default:
            logger.warning(
                "{!r} was set to {!r} in the environment. "
                "Changing the default allocation upon user request".format(
                    self._envvar, value
                )
            )
        return value


class Multipass(Provider):
    """A multipass provider for snapcraft to execute its lifecycle."""

    @classmethod
    def ensure_provider(cls):
        MultipassCommand.ensure_multipass(platform=sys.platform)

    @classmethod
    def setup_provider(cls, *, echoer) -> None:
        MultipassCommand.setup_multipass(echoer=echoer, platform=sys.platform)

    @classmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        return True

    @classmethod
    def _get_provider_name(cls):
        return "multipass"

    def _run(
        self, command: Sequence[str], hide_output: bool = False
    ) -> Optional[bytes]:
        has_tty = "SNAPCRAFT_HAS_TTY={}".format(sys.stdout.isatty())
        command = ["sudo", "-i", "env", has_tty] + list(command)
        return self._multipass_cmd.execute(
            instance_name=self.instance_name, command=command, hide_output=hide_output
        )

    def _get_disk_image(self) -> str:
        return "snapcraft:{}".format(self.project.info.get_build_base())

    def _launch(self) -> None:
        cloud_user_data_filepath = self._get_cloud_user_data()
        image = self._get_disk_image()

        cpus = _MachineSetting(envvar="SNAPCRAFT_BUILD_ENVIRONMENT_CPU", default="2")
        mem = _MachineSetting(envvar="SNAPCRAFT_BUILD_ENVIRONMENT_MEMORY", default="2G")
        disk = _MachineSetting(
            envvar="SNAPCRAFT_BUILD_ENVIRONMENT_DISK", default="256G"
        )

        self._multipass_cmd.launch(
            instance_name=self.instance_name,
            cpus=cpus.get_value(),
            mem=mem.get_value(),
            disk=disk.get_value(),
            image=image,
            cloud_init=cloud_user_data_filepath,
        )

    def _start(self):
        try:
            self._get_instance_info()
        except errors.ProviderInfoError as instance_error:
            # Until we have proper multipass error codes to know if this
            # was a communication error we should keep this error tracking
            # and generation here.
            raise errors.ProviderInstanceNotFoundError(
                instance_name=self.instance_name
            ) from instance_error

        self._multipass_cmd.start(instance_name=self.instance_name)

    def _mount(
        self,
        *,
        mountpoint: str,
        dev_or_path: str,
        uid_map: Dict[str, str] = None,
        gid_map: Dict[str, str] = None
    ) -> None:
        target = "{}:{}".format(self.instance_name, mountpoint)
        self._multipass_cmd.mount(
            source=dev_or_path, target=target, uid_map=uid_map, gid_map=gid_map
        )

    def _umount(self, *, mountpoint: str) -> None:
        mount = "{}:{}".format(self.instance_name, mountpoint)
        self._multipass_cmd.umount(mount=mount)

    def _push_file(self, *, source: str, destination: str) -> None:
        destination = "{}:{}".format(self.instance_name, destination)
        self._multipass_cmd.copy_files(source=source, destination=destination)

    def __init__(self, *, project, echoer, is_ephemeral: bool = False) -> None:
        super().__init__(project=project, echoer=echoer, is_ephemeral=is_ephemeral)
        self._multipass_cmd = MultipassCommand(platform=sys.platform)
        self._instance_info = None  # type: InstanceInfo

    def create(self) -> None:
        """Create the multipass instance and setup the build environment."""
        self.echoer.info("Launching a VM.")
        self.launch_instance()
        self._instance_info = self._get_instance_info()

    def destroy(self) -> None:
        """Destroy the instance, trying to stop it first."""
        try:
            instance_info = self._instance_info = self._get_instance_info()
        except errors.ProviderInfoError:
            return

        if instance_info.is_stopped():
            return

        stop_time = _get_stop_time()
        if stop_time > 0:
            try:
                self._multipass_cmd.stop(
                    instance_name=self.instance_name, time=stop_time
                )
            except errors.ProviderStopError:
                self._multipass_cmd.stop(instance_name=self.instance_name)
        else:
            self._multipass_cmd.stop(instance_name=self.instance_name)

        if self._is_ephemeral:
            self.clean_project()

    def mount_project(self) -> None:
        # Resolve the home directory
        home_dir = (
            self._run(command=["printenv", "HOME"], hide_output=True).decode().strip()
        )
        project_mountpoint = os.path.join(home_dir, "project")

        # multipass keeps the mount active, so check if it is there first.
        if not self._instance_info.is_mounted(project_mountpoint):
            self._mount(
                mountpoint=project_mountpoint,
                dev_or_path=self.project._project_dir,
                uid_map={str(os.getuid()): "0"},
                gid_map={str(os.getgid()): "0"},
            )

    def _mount_prime_directory(self) -> bool:
        # Resolve the home directory
        home_dir = (
            self._run(command=["printenv", "HOME"], hide_output=True).decode().strip()
        )
        prime_mountpoint = os.path.join(home_dir, "prime")
        if self._instance_info.is_mounted(prime_mountpoint):
            return True

        self._mount(
            mountpoint=prime_mountpoint,
            dev_or_path=self.project.prime_dir,
            uid_map={str(os.getuid()): "0"},
            gid_map={str(os.getgid()): "0"},
        )

        return False

    def clean_project(self) -> bool:
        was_cleaned = super().clean_project()
        if was_cleaned:
            self._multipass_cmd.delete(instance_name=self.instance_name, purge=True)
        return was_cleaned

    def pull_file(self, name: str, destination: str, delete: bool = False) -> None:
        # TODO add instance check.

        # check if file exists in instance
        self._run(command=["test", "-f", name])

        # copy file from instance
        source = "{}:{}".format(self.instance_name, name)
        self._multipass_cmd.copy_files(source=source, destination=destination)
        if delete:
            self._run(command=["rm", name])

    def shell(self) -> None:
        self._run(command=["/bin/bash"])

    def _get_instance_info(self):
        instance_info_raw = self._multipass_cmd.info(
            instance_name=self.instance_name, output_format="json"
        )
        return InstanceInfo.from_json(
            instance_name=self.instance_name, json_info=instance_info_raw.decode()
        )



import json
import os
import urllib.parse

import requests
from simplejson.scanner import JSONDecodeError

from ._client import Client
from ._status_tracker import StatusTracker

from . import _metadata
from . import constants
from . import errors
from . import logger, _macaroon_auth


class SCAClient(Client):
    """The software center agent deals with managing snaps."""

    def __init__(self, conf):
        super().__init__(
            conf,
            os.environ.get(
                "UBUNTU_STORE_API_ROOT_URL", constants.UBUNTU_STORE_API_ROOT_URL
            ),
        )

    def get_macaroon(self, acls, packages=None, channels=None, expires=None):
        data = {"permissions": acls}
        if packages is not None:
            data.update({"packages": packages})
        if channels is not None:
            data.update({"channels": channels})
        if expires is not None:
            data.update({"expires": expires})
        headers = {"Accept": "application/json"}
        response = self.post("acl/", json=data, headers=headers)
        if response.ok:
            return response.json()["macaroon"]
        else:
            raise errors.StoreAuthenticationError("Failed to get macaroon", response)

    @staticmethod
    def _is_needs_refresh_response(response):
        return (
            response.status_code == requests.codes.unauthorized
            and response.headers.get("WWW-Authenticate") == "Macaroon needs_refresh=1"
        )

    def request(self, *args, **kwargs):
        response = super().request(*args, **kwargs)
        if self._is_needs_refresh_response(response):
            raise errors.StoreMacaroonNeedsRefreshError()
        return response

    def verify_acl(self):
        auth = _macaroon_auth(self.conf)
        response = self.post(
            "acl/verify/",
            json={"auth_data": {"authorization": auth}},
            headers={"Accept": "application/json"},
        )
        if response.ok:
            return response.json()
        else:
            raise errors.StoreAccountInformationError(response)

    def get_account_information(self):
        auth = _macaroon_auth(self.conf)
        response = self.get(
            "account", headers={"Authorization": auth, "Accept": "application/json"}
        )
        if response.ok:
            return response.json()
        else:
            raise errors.StoreAccountInformationError(response)

    def register_key(self, account_key_request):
        data = {"account_key_request": account_key_request}
        auth = _macaroon_auth(self.conf)
        response = self.post(
            "account/account-key",
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StoreKeyRegistrationError(response)

    def register(
        self, snap_name: str, *, is_private: bool, series: str, store_id: str
    ) -> None:
        auth = _macaroon_auth(self.conf)
        data = dict(snap_name=snap_name, is_private=is_private, series=series)
        if store_id is not None:
            data["store"] = store_id
        response = self.post(
            "register-name/",
            data=json.dumps(data),
            headers={"Authorization": auth, "Content-Type": "application/json"},
        )
        if not response.ok:
            raise errors.StoreRegistrationError(snap_name, response)

    def snap_push_precheck(self, snap_name):
        data = {"name": snap_name, "dry_run": True}
        auth = _macaroon_auth(self.conf)
        response = self.post(
            "snap-push/",
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StorePushError(data["name"], response)

    def snap_push_metadata(
        self,
        snap_name,
        updown_data,
        delta_format=None,
        delta_hash=None,
        source_hash=None,
        target_hash=None,
        built_at=None,
    ):
        data = {
            "name": snap_name,
            "series": constants.DEFAULT_SERIES,
            "updown_id": updown_data["upload_id"],
            "binary_filesize": updown_data["binary_filesize"],
            "source_uploaded": updown_data["source_uploaded"],
        }
        if delta_format:
            data["delta_format"] = delta_format
            data["delta_hash"] = delta_hash
            data["source_hash"] = source_hash
            data["target_hash"] = target_hash
        if built_at is not None:
            data["built_at"] = built_at
        auth = _macaroon_auth(self.conf)
        response = self.post(
            "snap-push/",
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StorePushError(data["name"], response)

        return StatusTracker(response.json()["status_details_url"])

    def push_metadata(self, snap_id, snap_name, metadata, force):
        """Push the metadata to SCA."""
        metadata_handler = _metadata.StoreMetadataHandler(
            self, _macaroon_auth(self.conf), snap_id, snap_name
        )
        metadata_handler.push(metadata, force)

    def push_binary_metadata(self, snap_id, snap_name, metadata, force):
        """Push the binary metadata to SCA."""
        metadata_handler = _metadata.StoreMetadataHandler(
            self, _macaroon_auth(self.conf), snap_id, snap_name
        )
        metadata_handler.push_binary(metadata, force)

    def snap_release(self, snap_name, revision, channels, delta_format=None):
        data = {"name": snap_name, "revision": str(revision), "channels": channels}
        if delta_format:
            data["delta_format"] = delta_format
        auth = _macaroon_auth(self.conf)
        response = self.post(
            "snap-release/",
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StoreReleaseError(data["name"], response)

        response_json = response.json()

        return response_json

    def push_assertion(self, snap_id, assertion, endpoint, force):
        if endpoint == "validations":
            data = {"assertion": assertion.decode("utf-8")}
        elif endpoint == "developers":
            data = {"snap_developer": assertion.decode("utf-8")}
        auth = _macaroon_auth(self.conf)

        url = "snaps/{}/{}".format(snap_id, endpoint)

        # For `snap-developer`, revoking developers will require their uploads
        # to be invalidated.
        if force:
            url = url + "?ignore_revoked_uploads"

        response = self.put(
            url,
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )

        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = (
                "Invalid response from the server when pushing validations: {} {}"
            ).format(response.status_code, response)
            logger.debug(message)
            raise errors.StoreValidationError(
                snap_id, response, message="Invalid response from the server"
            )

        return response_json

    def get_assertion(self, snap_id, endpoint):
        auth = _macaroon_auth(self.conf)
        response = self.get(
            "snaps/{}/{}".format(snap_id, endpoint),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = "Invalid response from the server when getting {}: {} {}".format(
                endpoint, response.status_code, response
            )
            logger.debug(message)
            raise errors.StoreValidationError(
                snap_id, response, message="Invalid response from the server"
            )

        return response_json

    def push_snap_build(self, snap_id, snap_build):
        url = "snaps/{}/builds".format(snap_id)
        data = json.dumps({"assertion": snap_build})
        headers = {
            "Authorization": _macaroon_auth(self.conf),
            "Content-Type": "application/json",
        }
        response = self.post(url, data=data, headers=headers)
        if not response.ok:
            raise errors.StoreSnapBuildError(response)

    def snap_revisions(self, snap_id, series, arch):
        qs = {}
        if series:
            qs["series"] = series
        if arch:
            qs["arch"] = arch
        url = "snaps/" + snap_id + "/history"
        if qs:
            url += "?" + urllib.parse.urlencode(qs)
        auth = _macaroon_auth(self.conf)
        response = self.get(
            url,
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StoreSnapRevisionsError(response, snap_id, series, arch)

        response_json = response.json()

        return response_json

    def snap_status(self, snap_id, series, arch):
        qs = {}
        if series:
            qs["series"] = series
        if arch:
            qs["architecture"] = arch
        url = "snaps/" + snap_id + "/state"
        if qs:
            url += "?" + urllib.parse.urlencode(qs)
        auth = _macaroon_auth(self.conf)
        response = self.get(
            url,
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )
        if not response.ok:
            raise errors.StoreSnapStatusError(response, snap_id, series, arch)

        response_json = response.json()

        return response_json

    def close_channels(self, snap_id, channel_names):
        url = "snaps/{}/close".format(snap_id)
        data = {"channels": channel_names}
        headers = {"Authorization": _macaroon_auth(self.conf)}
        response = self.post(url, json=data, headers=headers)
        if not response.ok:
            raise errors.StoreChannelClosingError(response)

        try:
            results = response.json()
            return results["closed_channels"], results["channel_map_tree"]
        except (JSONDecodeError, KeyError):
            logger.debug(
                "Invalid response from the server on channel closing:\n"
                "{} {}\n{}".format(
                    response.status_code, response.reason, response.content
                )
            )
            raise errors.StoreChannelClosingError(response)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        auth = _macaroon_auth(self.conf)
        data = {"latest_tos_accepted": latest_tos_accepted}
        response = self.post(
            "agreement/",
            data=json.dumps(data),
            headers={
                "Authorization": auth,
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        )

        if not response.ok:
            raise errors.DeveloperAgreementSignError(response)
        return response.json()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import click

from . import env
from snapcraft.project import Project, get_snapcraft_yaml
from snapcraft.cli.echo import confirm, prompt
from snapcraft.internal import errors


class PromptOption(click.Option):
    def prompt_for_value(self, ctx):
        default = self.get_default(ctx)

        # If this is a prompt for a flag we need to handle this
        # differently.
        if self.is_bool_flag:
            return confirm(self.prompt, default)

        return prompt(
            self.prompt,
            default=default,
            type=self.type,
            hide_input=self.hide_input,
            show_choices=self.show_choices,
            confirmation_prompt=self.confirmation_prompt,
            value_proc=lambda x: self.process_value(ctx, x),
        )


_BUILD_OPTIONS = [
    dict(
        param_decls="--target-arch",
        metavar="<arch>",
        help="Target architecture to cross compile to",
    ),
    dict(
        param_decls="--debug",
        is_flag=True,
        help="Shells into the environment if the build fails.",
    ),
    dict(
        param_decls="--shell",
        is_flag=True,
        help="Shells into the environment in lieu of the step to run.",
    ),
    dict(
        param_decls="--shell-after",
        is_flag=True,
        help="Shells into the environment after the step has run.",
    ),
    dict(
        param_decls="--destructive-mode",
        is_flag=True,
        help="Forces snapcraft to try and use the current host to build.",
    ),
    dict(
        param_decls="--use-lxd",
        is_flag=True,
        help="Forces snapcraft to use LXD to build.",
    ),
]


def add_build_options(hidden=False):
    def _add_build_options(func):
        for build_option in reversed(_BUILD_OPTIONS):
            # Pop param_decls option to prevent exception further on down the
            # line for: `got multiple values for keyword argument`.
            build_option = build_option.copy()
            param_decls = build_option.pop("param_decls")
            option = click.option(param_decls, **build_option, hidden=hidden)
            func = option(func)
        return func

    return _add_build_options


def get_build_environment(**kwargs):
    force_use_lxd = kwargs.get("use_lxd")
    force_destructive_mode = kwargs.get("destructive_mode")

    if force_destructive_mode and force_use_lxd:
        raise click.BadOptionUsage(
            "--destructive-mode",
            "--use-lxd and --destructive-mode cannot be used together.",
        )
    elif force_use_lxd:
        provider = "lxd"
    elif force_destructive_mode:
        provider = "host"
    else:
        provider = None
    return env.BuilderEnvironmentConfig(force_provider=provider)


def get_project(*, is_managed_host: bool = False, **kwargs):
    # We need to do this here until we can get_snapcraft_yaml as part of Project.
    if is_managed_host:
        try:
            os.chdir(os.path.expanduser(os.path.join("~", "project")))
        except FileNotFoundError:
            # No project found (fresh environment).
            raise errors.ProjectNotFoundError()

    snapcraft_yaml_file_path = get_snapcraft_yaml()

    # This method may be called from a click.Command with no parent.
    ctx = click.get_current_context()
    if ctx.parent is not None:
        for key, value in ctx.parent.params.items():
            if not kwargs.get(key):
                kwargs[key] = value

    project = Project(
        debug=kwargs.pop("debug", False),
        target_deb_arch=kwargs.pop("target_arch", None),
        snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        is_managed_host=is_managed_host,
    )
    return project



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The nodejs plugin is useful for node/npm based parts.

The plugin uses node to install dependencies from `package.json`. It
also sets up binaries defined in `package.json` into the `PATH`.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - nodejs-version:
      (string)
      The version of nodejs you want the snap to run on.
      This includes npm, as would be downloaded from https://nodejs.org
      Defaults to the current LTS release.

    - nodejs-package-manager
      (string; default: yarn)
      The language package manager to use to drive installation
      of node packages. Can be either `npm` or `yarn` (default).

    - nodejs-yarn-version:
      (string)
      Applicable when using yarn. Defaults to the latest if not set.
"""

import collections
import contextlib
import json
import os
import shutil
import subprocess
import sys

import snapcraft
from snapcraft import sources
from snapcraft.internal import errors
from snapcraft.file_utils import link_or_copy, link_or_copy_tree


_NODEJS_BASE = "node-v{version}-linux-{arch}"
_NODEJS_VERSION = "8.12.0"
_NODEJS_TMPL = "https://nodejs.org/dist/v{version}/{base}.tar.gz"
_NODEJS_ARCHES = {
    "i386": "x86",
    "amd64": "x64",
    "armhf": "armv7l",
    "arm64": "arm64",
    "ppc64el": "ppc64le",
    "s390x": "s390x",
}
_YARN_LATEST_URL = "https://yarnpkg.com/latest.tar.gz"
# e.g.; https://github.com/yarnpkg/yarn/releases/download/v1.12.0/yarn-v1.12.0.tar.gz
_YARN_VERSION_URL = (
    "https://github.com/yarnpkg/yarn/releases/download/{version}/yarn-{version}.tar.gz"
)


class NodejsPluginMissingPackageJsonError(errors.SnapcraftError):

    fmt = (
        "Could not find a 'package.json' in the source tree.\n"
        "Verify that a 'package.json' exists at the root of the source tree\n"
        "or consider using the 'source-subdir' property if it is located in a subdirectory."
    )


class NodePlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["nodejs-version"] = {
            "type": "string",
            "default": _NODEJS_VERSION,
        }
        schema["properties"]["nodejs-package-manager"] = {
            "type": "string",
            "default": "yarn",
            "enum": ["npm", "yarn"],
        }
        schema["properties"]["nodejs-yarn-version"] = {"type": "string", "default": ""}

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["nodejs-version", "nodejs-package-manager", "nodejs-yarn-version"]

    @property
    def _nodejs_tar(self):
        if self._nodejs_tar_handle is None:
            self._nodejs_tar_handle = sources.Tar(
                self._nodejs_release_uri, self._npm_dir
            )
        return self._nodejs_tar_handle

    @property
    def _yarn_tar(self):
        if self._yarn_tar_handle is None:
            if self.options.nodejs_yarn_version:
                url = _YARN_VERSION_URL.format(version=self.options.nodejs_yarn_version)
            else:
                url = _YARN_LATEST_URL
            self._yarn_tar_handle = sources.Tar(url, self._npm_dir)
        return self._yarn_tar_handle

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._npm_dir = os.path.join(self.partdir, "npm")

        self._manifest = collections.OrderedDict()

        self._nodejs_release_uri = get_nodejs_release(
            self.options.nodejs_version, self.project.deb_arch
        )
        self._nodejs_tar_handle = None
        self._yarn_tar_handle = None

    def pull(self):
        super().pull()

        os.makedirs(self._npm_dir, exist_ok=True)
        self._nodejs_tar.download()
        if self.options.nodejs_package_manager == "yarn":
            self._yarn_tar.download()

        # do the install in the pull phase to download all dependencies.
        self._install(rootdir=self.sourcedir)

    def clean_pull(self):
        super().clean_pull()
        # Remove the npm directory (if any)
        if os.path.exists(self._npm_dir):
            shutil.rmtree(self._npm_dir)

    def build(self):
        super().build()

        package_dir = self._install(rootdir=self.builddir)

        # Now move everything over to the plugin's installdir
        link_or_copy_tree(package_dir, self.installdir)
        # Copy in the node binary
        link_or_copy(
            os.path.join(self._npm_dir, "bin", "node"),
            os.path.join(self.installdir, "bin", "node"),
        )
        # Create binary entries
        package_json = self._get_package_json(rootdir=self.builddir)
        _create_bins(package_json, self.installdir)

        lock_file_path = os.path.join(self.installdir, "yarn.lock")
        if os.path.isfile(lock_file_path):
            with open(lock_file_path) as lock_file:
                self._manifest["yarn-lock-contents"] = lock_file.read()

        # Get the names and versions of installed packages
        installed_node_packages = self._get_installed_node_packages(self.installdir)
        self._manifest["node-packages"] = [
            "{}={}".format(name, installed_node_packages[name])
            for name in installed_node_packages
        ]

    def _install(self, rootdir):
        self._nodejs_tar.provision(self._npm_dir, clean_target=False, keep_tarball=True)
        if self.options.nodejs_package_manager == "yarn":
            self._yarn_tar.provision(
                self._npm_dir, clean_target=False, keep_tarball=True
            )

        cmd = [os.path.join(self._npm_dir, "bin", self.options.nodejs_package_manager)]

        if self.options.nodejs_package_manager == "yarn":
            if os.getenv("http_proxy"):
                cmd.extend(["--proxy", os.getenv("http_proxy")])
            if os.getenv("https_proxy"):
                cmd.extend(["--https-proxy", os.getenv("https_proxy")])

        flags = []
        if rootdir == self.builddir:
            flags = ["--offline", "--prod"]

        self.run(cmd + ["install"] + flags, rootdir)

        package_json = self._get_package_json(rootdir)
        # Take into account scoped names
        name = package_json["name"].lstrip("@").replace("/", "-")
        version = package_json["version"]

        # npm pack will create a tarball of the form
        # <package-name>-<package-version>.tgz
        # and we will tell yarn pack to create one with this
        # predictable name.
        package_tar_path = "{name}-{version}.tgz".format(name=name, version=version)
        if self.options.nodejs_package_manager == "yarn":
            self.run(cmd + ["pack", "--filename", package_tar_path], rootdir)
        else:
            self.run(cmd + ["pack"], rootdir)

        package_dir = os.path.join(rootdir, "package")
        package_tar = sources.Tar(package_tar_path, rootdir)
        package_tar.file = package_tar_path
        os.makedirs(package_dir, exist_ok=True)
        package_tar.provision(package_dir)

        with contextlib.suppress(FileNotFoundError):
            shutil.copy(
                os.path.join(rootdir, "yarn.lock"),
                os.path.join(package_dir, "yarn.lock"),
            )

        self.run(cmd + ["install"] + flags, package_dir)

        return package_dir

    def run(self, cmd, rootdir):
        super().run(cmd, cwd=rootdir, env=self._build_environment())

    def run_output(self, cmd, rootdir):
        return super().run_output(cmd, cwd=rootdir, env=self._build_environment())

    def _build_environment(self):
        env = os.environ.copy()
        npm_bin = os.path.join(self._npm_dir, "bin")

        if env.get("PATH"):
            new_path = "{}:{}".format(npm_bin, env.get("PATH"))
        else:
            new_path = npm_bin

        env["PATH"] = new_path
        return env

    def _get_package_json(self, rootdir):
        try:
            with open(os.path.join(rootdir, "package.json")) as json_file:
                return json.load(json_file)
        except FileNotFoundError as not_found_error:
            raise NodejsPluginMissingPackageJsonError() from not_found_error

    def _get_installed_node_packages(self, cwd):
        # There is no yarn ls
        cmd = [os.path.join(self._npm_dir, "bin", "npm"), "ls", "--json"]
        try:
            output = self.run_output(cmd, cwd)
        except subprocess.CalledProcessError as error:
            # XXX When dependencies have missing dependencies, an error like
            # this is printed to stderr:
            # npm ERR! peer dep missing: glob@*, required by glob-promise@3.1.0
            # retcode is not 0, which raises an exception.
            output = error.output.decode(sys.getfilesystemencoding()).strip()
        packages = collections.OrderedDict()
        output_json = json.loads(output, object_pairs_hook=collections.OrderedDict)
        dependencies = output_json.get("dependencies", [])
        while dependencies:
            key, value = dependencies.popitem(last=False)
            # XXX Just as above, dependencies without version are the ones
            # missing.
            if "version" in value:
                packages[key] = value["version"]
            if "dependencies" in value:
                dependencies.update(value["dependencies"])
        return packages

    def get_manifest(self):
        return self._manifest


def _create_bins(package_json, directory):
    bin_entry = package_json.get("bin")
    if not bin_entry:
        return

    bin_dir = os.path.join(directory, "bin")
    os.makedirs(bin_dir, exist_ok=True)

    if type(bin_entry) == dict:
        binaries = bin_entry
    elif type(bin_entry) == str:
        # Support for scoped names of the form of @org/name
        name = package_json["name"]
        binaries = {name[name.find("/") + 1 :]: bin_entry}
    else:
        raise errors.SnapcraftEnvironmentError(
            "The plugin is not prepared to handle bin entries of "
            "type {!r}".format(type(bin_entry))
        )

    for bin_name, bin_path in binaries.items():
        target = os.path.join(bin_dir, bin_name)
        # The binary might be already created from upstream sources.
        if os.path.exists(os.path.join(target)):
            continue
        source = os.path.join("..", bin_path)
        os.symlink(source, target)
        # Make it executable
        os.chmod(os.path.realpath(target), 0o755)


def _get_nodejs_base(node_engine, machine):
    if machine not in _NODEJS_ARCHES:
        raise errors.SnapcraftEnvironmentError(
            "architecture not supported ({})".format(machine)
        )
    return _NODEJS_BASE.format(version=node_engine, arch=_NODEJS_ARCHES[machine])


def get_nodejs_release(node_engine, arch):
    return _NODEJS_TMPL.format(
        version=node_engine, base=_get_nodejs_base(node_engine, arch)
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import tempfile

from . import errors
from ._base import FileBase
from snapcraft import file_utils, yaml_utils


class Snap(FileBase):
    """Handles downloading and extractions for a snap source.

    On provision, the meta directory is renamed to meta.<snap-name> and, if present,
    the same applies for the snap directory which shall be renamed to
    snap.<snap-name>.
    """

    def __init__(
        self,
        source: str,
        source_dir: str,
        source_tag: str = None,
        source_commit: str = None,
        source_branch: str = None,
        source_depth: str = None,
        source_checksum: str = None,
    ) -> None:
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
            "unsquashfs",
        )
        if source_tag:
            raise errors.SnapcraftSourceInvalidOptionError("snap", "source-tag")
        elif source_commit:
            raise errors.SnapcraftSourceInvalidOptionError("snap", "source-commit")
        elif source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("snap", "source-branch")

    def provision(
        self,
        dst: str,
        clean_target: bool = True,
        keep_snap: bool = False,
        src: str = None,
    ) -> None:
        """
        Provision the snap source to dst.

        :param str dst: the destination directory to provision to.
        :param bool clean_target: clean dst before provisioning if True.
        :param bool keep_snap: keep the snap after provisioning is done if
                               True.
        :param str src: force a new source to use for extraction.
        raises errors.InvalidSnapError: when trying to provision an invalid
                                        snap.
        """
        if src:
            snap_file = src
        else:
            snap_file = os.path.join(self.source_dir, os.path.basename(self.source))
        snap_file = os.path.realpath(snap_file)

        if clean_target:
            tmp_snap = tempfile.NamedTemporaryFile().name
            shutil.move(snap_file, tmp_snap)
            shutil.rmtree(dst)
            os.makedirs(dst)
            shutil.move(tmp_snap, snap_file)

        # unsquashfs [options] filesystem [directories or files to extract]
        # options:
        # -force: if file already exists then overwrite
        # -dest <pathname>: unsquash to <pathname>
        with tempfile.TemporaryDirectory(prefix=os.path.dirname(snap_file)) as temp_dir:
            extract_command = ["unsquashfs", "-force", "-dest", temp_dir, snap_file]
            self._run_output(extract_command)
            snap_name = _get_snap_name(temp_dir)
            # Rename meta and snap dirs from the snap
            rename_paths = (os.path.join(temp_dir, d) for d in ["meta", "snap"])
            rename_paths = (d for d in rename_paths if os.path.exists(d))
            for rename in rename_paths:
                shutil.move(rename, "{}.{}".format(rename, snap_name))
            file_utils.link_or_copy_tree(source_tree=temp_dir, destination_tree=dst)

        if not keep_snap:
            os.remove(snap_file)


def _get_snap_name(snap_dir: str) -> str:
    try:
        with open(os.path.join(snap_dir, "meta", "snap.yaml")) as snap_yaml:
            return yaml_utils.load(snap_yaml)["name"]
    except (FileNotFoundError, KeyError) as snap_error:
        raise errors.InvalidSnapError() from snap_error



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Any, Dict, List, Optional

from . import errors
from snapcraft.file_utils import calculate_hash


"""
This module holds representations for results from the v2 info API provided by
the Snap Store.

The full API is documented on https://api.snapcraft.io/api-docs/info.html
"""

# FIXME: attributes are manually created to keep the version of mypy we are using happy.
#        This might be a good thing in the long run.


class SnapChannelDetails:
    def __repr__(self) -> str:
        return "<SnapChannelDetails: {!r} for arch {!r} released at {!r}>".format(
            self.name, self.architecture, self.released_at
        )

    def __init__(self, details: Dict[str, Any]) -> None:
        self._payload = details

    @property
    def architecture(self) -> str:
        return self._payload["architecture"]

    @property
    def name(self) -> str:
        return self._payload["name"]

    @property
    def released_at(self) -> str:
        return self._payload["released-at"]

    @property
    def risk(self) -> str:
        return self._payload["risk"]

    @property
    def track(self) -> str:
        return self._payload["track"]


class SnapDownloadDetails:
    def __repr__(self) -> str:
        return "<SnapDownloadDetails: {!r} with size {!r}>".format(self.url, self.size)

    def __init__(self, download: Dict[str, Any]) -> None:
        self._payload = download

    @property
    def url(self) -> str:
        return self._payload["url"]

    @property
    def size(self) -> int:
        return self._payload["size"]

    @property
    def sha3_384(self) -> str:
        return self._payload["sha3-384"]

    def verify(self, path: str) -> None:
        if not os.path.exists(path):
            raise errors.DownloadNotFoundError(path=path)

        calculated_hash = calculate_hash(path, algorithm="sha3_384")
        if self.sha3_384 != calculated_hash:
            raise errors.SHAMismatchError(
                path=path, expected=self.sha3_384, calculated=calculated_hash
            )


class SnapPublisherDetails:
    def __repr__(self) -> str:
        return "<SnapPublisherDetails: {!r}>".format(self.username)

    def __init__(self, publisher: Dict[str, Any]) -> None:
        self._payload = publisher

    @property
    def id(self) -> str:
        return self._payload["id"]

    @property
    def display_name(self) -> str:
        return self._payload["display-name"]

    @property
    def username(self) -> str:
        return self._payload["username"]

    @property
    def validation(self) -> str:
        return self._payload["validation"]


class SnapDetails:
    def __repr__(self) -> str:
        return "<SnapDetails: {!r} with snap-id {!r}>".format(self.name, self.snap_id)

    def __init__(self, snap: Dict[str, Any]) -> None:
        self._payload = snap
        self._publisher = None  # type: Optional[SnapPublisherDetails]

    @property
    def publisher(self) -> SnapPublisherDetails:
        if self._publisher is None:
            self._publisher = SnapPublisherDetails(self._payload.get("publisher"))
        return self._publisher

    @property
    def name(self) -> str:
        return self._payload["name"]

    @property
    def snap_id(self) -> str:
        return self._payload["snap-id"]


class SnapChannelMapping:
    def __repr__(self) -> str:
        return "<SnapChannelMapping: {!r} on {!r}>".format(
            self.channel_details, self.revision
        )

    def __init__(self, channel: Dict[str, Any]) -> None:
        self._payload = channel
        self._channel_details = None  # type: Optional[SnapChannelDetails]
        self._download = None  # type: Optional[SnapDownloadDetails]

    @property
    def channel_details(self) -> SnapChannelDetails:
        if self._channel_details is None:
            self._channel_details = SnapChannelDetails(self._payload.get("channel"))
        return self._channel_details

    @property
    def download(self) -> SnapDownloadDetails:
        if self._download is None:
            self._download = SnapDownloadDetails(self._payload.get("download"))
        return self._download

    @property
    def revision(self) -> int:
        return self._payload["revision"]

    @property
    def confinement(self) -> str:
        return self._payload["confinement"]

    @property
    def version(self) -> str:
        return self._payload["version"]


class SnapInfo:
    def __repr__(self) -> str:
        return "<SnapInfo: {!r}>".format(self.name)

    def __init__(self, snap_info_resp: Dict[str, Any]) -> None:
        self._payload = snap_info_resp
        self._channel_map = None  # type: Optional[List[SnapChannelMapping]]
        self._snap = None  # type: Optional[SnapDetails]

    @property
    def channel_map(self) -> List[SnapChannelMapping]:
        if self._channel_map is None:
            self._channel_map = [
                SnapChannelMapping(i) for i in self._payload.get("channel-map", [])
            ]
        return self._channel_map

    @property
    def snap(self) -> SnapDetails:
        if self._snap is None:
            self._snap = SnapDetails(self._payload.get("snap"))
        return self._snap

    @property
    def name(self) -> str:
        return self._payload["name"]

    @property
    def snap_id(self) -> str:
        return self._payload["snap-id"]

    def get_channel_mapping(
        self, *, risk: str, arch: str, track: Optional[str] = None
    ) -> SnapChannelMapping:
        if track is None:
            track_filter = "latest"
        else:
            track_filter = track

        arch_match = (
            c for c in self.channel_map if c.channel_details.architecture == arch
        )
        track_match = (c for c in arch_match if c.channel_details.track == track_filter)
        risk_match = [c for c in track_match if c.channel_details.risk == risk]

        if not risk_match:
            channel = "{}/{}".format(track, risk) if track else risk
            raise errors.SnapNotFoundError(name=self.name, channel=channel)

        # We assume the API will not return duplicate mappings
        return risk_match[0]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Manas.Tech
# License granted by Canonical Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The Crystal plugin can be used for Crystal projects using `shards`.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - crystal-channel:
      (string, default: latest/stable)
      The Snap Store channel to install Crystal from.
"""

import os
import shutil

import snapcraft
from snapcraft import file_utils
from snapcraft.internal import common, elf, errors

_CRYSTAL_CHANNEL = "latest/stable"


class CrystalPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["crystal-channel"] = {
            "type": "string",
            "default": _CRYSTAL_CHANNEL,
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["crystal-channel"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        if project.info.get_build_base() not in ("core", "core16", "core18"):
            raise errors.PluginBaseError(
                part_name=self.name, base=project.info.get_build_base()
            )

        self.build_snaps.append("crystal/{}".format(self.options.crystal_channel))
        self.build_packages.extend(
            [
                "gcc",
                "pkg-config",
                "libpcre3-dev",
                "libevent-dev",
                "libyaml-dev",
                "libgmp-dev",
                "libxml2-dev",
            ]
        )

    def build(self):
        super().build()

        self.run(["shards", "install", "--production"], self.builddir)
        self.run(["shards", "build", "--production"], self.builddir)

        output_bin = os.path.join(self.builddir, "bin")
        if not os.path.exists(output_bin):
            raise errors.SnapcraftEnvironmentError(
                "No binaries were built. Ensure the shards.yaml contains valid targets."
            )

        install_bin_path = os.path.join(self.installdir, "bin")

        bin_paths = (os.path.join(output_bin, b) for b in os.listdir(output_bin))
        elf_files = (elf.ElfFile(path=b) for b in bin_paths if elf.ElfFile.is_elf(b))

        os.makedirs(install_bin_path, exist_ok=True)

        for elf_file in elf_files:
            shutil.copy2(
                elf_file.path,
                os.path.join(install_bin_path, os.path.basename(elf_file.path)),
            )

            elf_dependencies_path = elf_file.load_dependencies(
                root_path=self.installdir,
                core_base_path=common.get_core_path(self.project.info.get_build_base()),
            )
            for elf_dependency_path in elf_dependencies_path:
                lib_install_path = os.path.join(
                    self.installdir, elf_dependency_path[1:]
                )
                os.makedirs(os.path.dirname(lib_install_path), exist_ok=True)
                if not os.path.exists(lib_install_path):
                    file_utils.link_or_copy(
                        elf_dependency_path, lib_install_path, follow_symlinks=True
                    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys

import click

from ._options import get_build_environment, get_project
from snapcraft.internal import common, errors


def run_legacy_snapcraft(argv=sys.argv[1:]) -> None:
    if not common.is_snap():
        raise errors.SnapcraftEnvironmentError(
            "Legacy mode not supported in this installation. "
            "Install snapcraft from https://snapcraft.io/snapcraft and try again."
        )

    legacy_python = os.path.join(
        common.get_legacy_snapcraft_dir(), "usr", "bin", "python3"
    )
    legacy_snapcraft = os.path.join(
        common.get_legacy_snapcraft_dir(), "bin", "snapcraft"
    )

    cmd = [legacy_python, legacy_snapcraft] + argv
    logging.debug("Running legacy snapcraft with: {}".format(cmd))
    os.execv(legacy_python, cmd)


class SnapcraftProjectCommand(click.Command):
    """
    Project related command that can fetch properties from the legacy tree.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _is_legacy(self) -> bool:
        try:
            build_environment = get_build_environment()
            project = get_project(is_managed_host=build_environment.is_managed_host)
        except Exception:
            # If we cannot load the project, we assume we are in non legacy
            # where the error should be properly handled.
            return False

        # Not in an actual project.
        if project is None or project.info is None:
            return False
        # Building a base is not legacy.
        elif project.info.type == "base":
            return False
        elif project.info.build_base is not None:
            return False
        # A project with no base set is legacy.
        elif project.info.base is None:
            return True
        else:
            return False

    def get_help(self, ctx):
        if self._is_legacy():
            run_legacy_snapcraft(argv=["help", self.name])
        else:
            return super().get_help(ctx)

    def invoke(self, ctx):
        if self._is_legacy():
            run_legacy_snapcraft()
        else:
            return super().invoke(ctx)



# -*- Mode:Python; indent-tabs-buildnil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Dict, List, Type

from snapcraft import yaml_utils
from snapcraft.internal.states._state import State


class GlobalState(State):

    yaml_tag = u"!GlobalState"

    @classmethod
    def load(cls: Type["GlobalState"], *, filepath: str) -> "GlobalState":
        with open(filepath) as state_file:
            return yaml_utils.load(state_file)

    def save(self, *, filepath: str) -> None:
        dirpath = os.path.dirname(filepath)
        if dirpath:
            os.makedirs(dirpath, exist_ok=True)
        with open(filepath, "w") as state_file:
            yaml_utils.dump(self, stream=state_file)

    def get_build_packages(self) -> List[str]:
        return self.assets.get("build-packages", [])

    def append_build_packages(self, build_packages: List[str]) -> None:
        current_build_packages = self.get_build_packages()
        new_packages = [b for b in build_packages if b not in current_build_packages]
        self.assets["build-packages"] = current_build_packages + new_packages

    def get_build_snaps(self) -> List[str]:
        return self.assets.get("build-snaps", [])

    def append_build_snaps(self, build_snaps: List[str]) -> None:
        current_build_snaps = self.get_build_snaps()
        new_snaps = [b for b in build_snaps if b not in current_build_snaps]
        self.assets["build-snaps"] = current_build_snaps + new_snaps

    def __init__(self, *, assets: Dict[str, List[str]] = None) -> None:
        super().__init__()
        if assets is None:
            self.assets = dict()  # type: Dict[str, List[str]]
        else:
            self.assets = assets



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from snapcraft.internal.errors import OsReleaseIdError
from snapcraft.internal.os_release import OsRelease

logger = logging.getLogger(__name__)

_DEB_BASED_PLATFORM = ["ubuntu", "debian", "elementary OS", "elementary", "neon"]


def _is_deb_based(distro=None):
    if not distro:
        try:
            distro = OsRelease().id()
        except OsReleaseIdError:
            distro = "unknown"
    return distro in _DEB_BASED_PLATFORM


def _get_repo_for_platform():
    if _is_deb_based():
        from ._deb import Ubuntu

        return Ubuntu
    else:
        from ._base import DummyRepo

        return DummyRepo



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import subprocess
import tempfile
from typing import List, Set

import snapcraft
from snapcraft.internal import common, repo
from snapcraft import formatting_utils

logger = logging.getLogger(__name__)


class Rospack:
    def __init__(
        self,
        *,
        ros_distro: str,
        ros_package_path: str,
        rospack_path: str,
        ubuntu_sources: str,
        ubuntu_keyrings: List[str],
        project: snapcraft.project.Project
    ) -> None:
        """Create a new Rospack instance.

        :param str ros_distro: Name of the ROS distribution.
        :param str ros_package_path: The workspace to use.
        :param str rospack_path: Working directory for rospack (where it will be
                                 installed).
        :param str ubuntu_sources: Ubuntu repositories from which rospack will be
                                   installed.
        :param list ubuntu_keyrings: List of paths to keyrings to use for sources.
        :param project: Instance of Project for project-wide settings.
        :type project: snapcraft.Project
        """
        self._ros_distro = ros_distro
        self._ros_package_path = ros_package_path
        self._rospack_path = rospack_path
        self._ubuntu_sources = ubuntu_sources
        self._ubuntu_keyrings = ubuntu_keyrings
        self._project = project

        self._rospack_install_path = os.path.join(self._rospack_path, "install")
        self._rospack_cache_path = os.path.join(self._rospack_path, "cache")

    def setup(self) -> None:
        """Fetch, unpack, and setup rospack."""

        # Make sure we can run multiple times without error
        os.makedirs(self._rospack_install_path, exist_ok=True)

        # rospack isn't necessarily a dependency of the project, so we'll unpack
        # it off to the side and use it from there.
        logger.info("Preparing to fetch rospack...")
        ubuntu_repo = repo.Ubuntu(
            self._rospack_path,
            sources=self._ubuntu_sources,
            keyrings=self._ubuntu_keyrings,
            project_options=self._project,
        )

        logger.info("Fetching rospack...")
        ubuntu_repo.get(
            [
                "ros-{}-rospack".format(self._ros_distro),
                "ros-{}-catkin".format(self._ros_distro),
            ]
        )

        logger.info("Installing rospack...")
        ubuntu_repo.unpack(self._rospack_install_path)

    def list_names(self) -> Set[str]:
        """Obtain list of packages present in the workspace."""

        output = self._run(["list-names"]).strip()
        if output:
            return set(output.split("\n"))
        else:
            return set()

    def _run(self, arguments) -> str:
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            lines = [
                'export PYTHONPATH="{}"'.format(
                    os.path.join(
                        self._rospack_install_path,
                        "usr",
                        "lib",
                        "python2.7",
                        "dist-packages",
                    )
                )
            ]

            ros_path = os.path.join(
                self._rospack_install_path, "opt", "ros", self._ros_distro
            )
            bin_paths = (
                os.path.join(ros_path, "bin"),
                os.path.join(self._rospack_install_path, "usr", "bin"),
            )
            lines.append(
                "export {}".format(
                    formatting_utils.format_path_variable(
                        "PATH", bin_paths, prepend="", separator=":"
                    )
                )
            )

            lib_paths = common.get_library_paths(
                self._rospack_install_path, self._project.arch_triplet
            )
            if lib_paths:
                lines.append(
                    "export {}".format(
                        formatting_utils.format_path_variable(
                            "LD_LIBRARY_PATH", lib_paths, prepend="", separator=":"
                        )
                    )
                )

            # Source our own workspace so we have all of rospack's dependencies
            lines.append(
                "_CATKIN_SETUP_DIR={} source {}".format(
                    ros_path, os.path.join(ros_path, "setup.sh")
                )
            )

            # By default, rospack saves its cache in $HOME/.ros, which we shouldn't
            # access here, so we'll redirect it with this environment variable.
            lines.append('export ROS_HOME="{}"'.format(self._rospack_cache_path))

            lines.append('export ROS_PACKAGE_PATH="{}"'.format(self._ros_package_path))

            lines.append('exec "$@"')
            f.write("\n".join(lines))
            f.flush()
            return (
                subprocess.check_output(
                    ["/bin/bash", f.name, "rospack"] + arguments,
                    stderr=subprocess.STDOUT,
                )
                .decode("utf8")
                .strip()
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.common import get_python2_path  # noqa
from snapcraft.internal.common import isurl  # noqa
from snapcraft.internal.common import get_include_paths  # noqa
from snapcraft.internal.common import get_library_paths  # noqa

# These are now available via file_utils, but don't break API.
from snapcraft.file_utils import link_or_copy  # noqa
from snapcraft.file_utils import replace_in_file  # noqa

# These are now available via formatting_utils, but don't break API.
from snapcraft.formatting_utils import combine_paths  # noqa
from snapcraft.formatting_utils import format_path_variable  # noqa



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from typing import Any, Dict, Type

from snapcraft.internal.build_providers import errors


class InstanceInfo:
    @classmethod
    def from_json(
        cls: Type["InstanceInfo"], *, instance_name: str, json_info: str
    ) -> "InstanceInfo":
        """Create an InstanceInfo from json_info retrieved from multipass.

        :param str instance_name: the name of the instance.
        :param str json_info: a json formatted string with the structure
                              that would follow the output of a json formatted
                              multipass info command.
        :returns: an InstanceInfo.
        :rtype: InstanceInfo
        :raises snapcraft.internal.build_providers.ProviderInfoDataKeyError:
            if the instance name cannot be found in the given json or if a
            required key is missing from that data structure for the instance.
        """
        try:
            json_data = json.loads(json_info)
        except json.decoder.JSONDecodeError as decode_error:
            raise errors.ProviderBadDataError(
                provider_name="multipass", data=json_info
            ) from decode_error
        try:
            instance_info = json_data["info"][instance_name]
        except KeyError as missing_key:
            raise errors.ProviderInfoDataKeyError(
                provider_name="multipass", missing_key=str(missing_key), data=json_data
            ) from missing_key
        try:
            return cls(
                name=instance_name,
                state=instance_info["state"],
                image_release=instance_info["image_release"],
                mounts=instance_info["mounts"],
            )
        except KeyError as missing_key:
            raise errors.ProviderInfoDataKeyError(
                provider_name="multipass",
                missing_key=str(missing_key),
                data=instance_info,
            ) from missing_key

    def __init__(
        self, *, name: str, state: str, image_release: str, mounts: Dict[str, Any]
    ) -> None:
        """Initialize an InstanceInfo.

        :param str name: the instance name.
        :param str state: the state of the instance which can be any one of
                          RUNNING, STOPPED, DELETED.
        :param str image_release: the Operating System release string for the
                                  image.
        """
        # We do not check for validity of state given that new states could
        # be introduced.
        self.name = name
        self.state = state
        self.image_release = image_release
        self.mounts = mounts

    def is_mounted(self, mountpoint: str) -> bool:
        return mountpoint in self.mounts

    def is_stopped(self) -> bool:
        return self.state == "STOPPED"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import functools
import glob
import hashlib
import logging
import os
import re
import shutil
import stat
import string
import subprocess
import sys
import urllib
import urllib.request
from typing import Dict, Set, List, Tuple  # noqa: F401

import apt
from xml.etree import ElementTree

import snapcraft
from snapcraft import file_utils
from snapcraft.internal import cache, repo, common, os_release
from snapcraft.internal.indicators import is_dumb_terminal
from ._base import BaseRepo
from . import errors


logger = logging.getLogger(__name__)

_DEFAULT_SOURCES = """deb http://${prefix}.ubuntu.com/${suffix}/ ${release} main restricted
deb http://${prefix}.ubuntu.com/${suffix}/ ${release}-updates main restricted
deb http://${prefix}.ubuntu.com/${suffix}/ ${release} universe
deb http://${prefix}.ubuntu.com/${suffix}/ ${release}-updates universe
deb http://${prefix}.ubuntu.com/${suffix}/ ${release} multiverse
deb http://${prefix}.ubuntu.com/${suffix}/ ${release}-updates multiverse
deb http://${security}.ubuntu.com/${suffix} ${release}-security main restricted
deb http://${security}.ubuntu.com/${suffix} ${release}-security universe
deb http://${security}.ubuntu.com/${suffix} ${release}-security multiverse
"""
_GEOIP_SERVER = "http://geoip.ubuntu.com/lookup"
_library_list = dict()  # type: Dict[str, Set[str]]
_HASHSUM_MISMATCH_PATTERN = re.compile(r"(E:Failed to fetch.+Hash Sum mismatch)+")


@functools.lru_cache(maxsize=256)
def _run_dpkg_query_s(file_path: str) -> str:
    try:
        output = (
            subprocess.check_output(
                ["dpkg-query", "-S", os.path.join(os.path.sep, file_path)],
                stderr=subprocess.STDOUT,
                env=dict(LANG="C.UTF-8"),
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError as call_error:
        logger.debug(
            "Error finding package for {}: {}".format(file_path, str(call_error))
        )
        raise errors.FileProviderNotFound(file_path=file_path) from call_error

    # Remove diversions
    provides_output = [p for p in output.splitlines() if not p.startswith("diversion")][
        0
    ]
    return provides_output.split(":")[0]


class _AptCache:
    def __init__(self, deb_arch, *, sources_list=None, keyrings=None, use_geoip=False):
        self._deb_arch = deb_arch
        self._sources_list = sources_list
        self._use_geoip = use_geoip

        if not keyrings:
            keyrings = list()
        self._keyrings = keyrings

    def _setup_apt(self, cache_dir):
        # Do not install recommends
        apt.apt_pkg.config.set("Apt::Install-Recommends", "False")

        # Ensure repos are provided by trusted third-parties
        apt.apt_pkg.config.set("Acquire::AllowInsecureRepositories", "False")

        # Methods and solvers dir for when in the SNAP
        if common.is_snap():
            snap_dir = os.getenv("SNAP")
            apt_dir = os.path.join(snap_dir, "usr", "lib", "apt")
            apt.apt_pkg.config.set("Dir", apt_dir)
            # yes apt is broken like that we need to append os.path.sep
            methods_dir = os.path.join(apt_dir, "methods")
            apt.apt_pkg.config.set("Dir::Bin::methods", methods_dir + os.path.sep)
            solvers_dir = os.path.join(apt_dir, "solvers")
            apt.apt_pkg.config.set("Dir::Bin::solvers::", solvers_dir + os.path.sep)
            apt_key_path = os.path.join(snap_dir, "usr", "bin", "apt-key")
            apt.apt_pkg.config.set("Dir::Bin::apt-key", apt_key_path)
            gpgv_path = os.path.join(snap_dir, "usr", "bin", "gpgv")
            apt.apt_pkg.config.set("Apt::Key::gpgvcommand", gpgv_path)
            apt.apt_pkg.config.set("Dir::Etc::Trusted", "/etc/apt/trusted.gpg")
            apt.apt_pkg.config.set("Dir::Etc::TrustedParts", "/etc/apt/trusted.gpg.d/")

        # Make sure we always use the system GPG configuration, even with
        # apt.Cache(rootdir). However, we also want to be able to add keys to it
        # without root, so symlink back to the system's, but maintain our own.
        # We'll leave Trusted alone and just fiddle with TrustedParts (Trusted is the
        # one modified by apt-key, so add-apt-repository should still work).
        apt.apt_pkg.config.set(
            "Dir::Etc::Trusted", apt.apt_pkg.config.find_file("Dir::Etc::Trusted")
        )
        apt_config_path = os.path.join(cache_dir, "etc", "apt", "apt.conf")
        trusted_parts_path = apt.apt_pkg.config.find_file("Dir::Etc::TrustedParts")
        if not trusted_parts_path.startswith(cache_dir):
            cached_trusted_parts = os.path.join(
                cache_dir, trusted_parts_path.lstrip("/")
            )
            with contextlib.suppress(FileNotFoundError):
                shutil.rmtree(cached_trusted_parts)
            os.makedirs(cached_trusted_parts)
            for trusted_part in os.scandir(trusted_parts_path):
                os.symlink(
                    os.path.join(trusted_parts_path, trusted_part.name),
                    os.path.join(cached_trusted_parts, trusted_part.name),
                )

            apt.apt_pkg.config.set("Dir::Etc::TrustedParts", cached_trusted_parts)

            # The above config is all that is needed on bionic, but xenial
            # requires this configuration file
            os.makedirs(os.path.dirname(apt_config_path), exist_ok=True)
            with open(apt_config_path, "w") as f:
                f.write("Dir::Etc::TrustedParts {};\n".format(cached_trusted_parts))

            # Now copy in any requested keyrings
            for keyring in self._keyrings:
                shutil.copy2(keyring, cached_trusted_parts)

        # Clear up apt's Post-Invoke-Success as we are not running
        # on the system.
        apt.apt_pkg.config.clear("APT::Update::Post-Invoke-Success")

        self.progress = apt.progress.text.AcquireProgress()
        if is_dumb_terminal():
            # Make output more suitable for logging.
            self.progress.pulse = lambda owner: True
            self.progress._width = 0

        sources_list_file = os.path.join(cache_dir, "etc", "apt", "sources.list")

        os.makedirs(os.path.dirname(sources_list_file), exist_ok=True)
        with open(sources_list_file, "w") as f:
            f.write(self._collected_sources_list())

        # dpkg also needs to be in the rootdir in order to support multiarch
        # (apt calls dpkg --print-foreign-architectures).
        dpkg_path = shutil.which("dpkg")
        if dpkg_path:
            # Symlink it into place
            destination = os.path.join(cache_dir, dpkg_path[1:])
            if not os.path.exists(destination):
                os.makedirs(os.path.dirname(destination), exist_ok=True)
                os.symlink(dpkg_path, destination)
        else:
            logger.warning("Cannot find 'dpkg' command needed to support multiarch")

        old_apt_config = os.environ.get("APT_CONFIG")
        try:
            # Only xenial needs this. If snapcraft changes to core18, this
            # variable is unnecessary.
            os.environ["APT_CONFIG"] = apt_config_path
            return self._create_cache(cache_dir, sources_list_file)
        finally:
            if old_apt_config is None:
                del os.environ["APT_CONFIG"]
            else:
                os.environ["APT_CONFIG"] = old_apt_config

    def _create_cache(self, cache_dir: str, sources_list_file: str) -> apt.Cache:
        apt_cache = apt.Cache(rootdir=cache_dir, memonly=True)
        try:
            apt_cache.update(
                fetch_progress=self.progress, sources_list=sources_list_file
            )
        except apt.cache.FetchFailedException as e:
            # In case of a hashsum mismatch, clear the index and try again.
            # Re-raise the error if that failed.
            if re.match(_HASHSUM_MISMATCH_PATTERN, str(e)):
                try:
                    index = apt.apt_pkg.config.find_dir("Dir::State::Lists")
                    shutil.rmtree(index)
                    apt_cache = apt.Cache(rootdir=cache_dir, memonly=True)
                    apt_cache.update(
                        fetch_progress=self.progress, sources_list=sources_list_file
                    )
                except apt.cache.FetchFailedException as retry:
                    raise errors.CacheUpdateFailedError(str(retry))
            else:
                raise errors.CacheUpdateFailedError(str(e))
        return apt_cache

    @contextlib.contextmanager
    def archive(self, cache_dir):
        try:
            apt_cache = self._setup_apt(cache_dir)
            apt_cache.open()

            try:
                yield apt_cache
            finally:
                apt_cache.close()
        except Exception as e:
            logger.debug("Exception occurred: {!r}".format(e))
            raise e

    def sources_digest(self):
        return hashlib.sha384(
            self._collected_sources_list().encode(sys.getfilesystemencoding())
        ).hexdigest()

    def _collected_sources_list(self):
        if self._use_geoip or self._sources_list:
            release = os_release.OsRelease()
            return _format_sources_list(
                self._sources_list,
                deb_arch=self._deb_arch,
                use_geoip=self._use_geoip,
                release=release.version_codename(),
            )

        return _get_local_sources_list()

    def fetch_binary(self, *, package_candidate, destination: str) -> None:
        # This is a workaround for the overly verbose python-apt we use.
        # There is an unreleased patch which once released could replace
        # this code https://salsa.debian.org/apt-team/python-apt/commit/d122f9142df614dbb5f7644112280140dc155ecc  # noqa
        # What follows is almost a tit for tat implementation of upstream's
        # fetch_binary logic.
        base = os.path.basename(package_candidate._records.filename)
        destfile = os.path.join(destination, base)
        if apt.package._file_is_same(
            destfile, package_candidate.size, package_candidate._records.md5_hash
        ):
            logging.debug("Ignoring already existing file: {}".format(destfile))
            return os.path.abspath(destfile)
        acq = apt.apt_pkg.Acquire(self.progress)
        acqfile = apt.apt_pkg.AcquireFile(
            acq,
            package_candidate.uri,
            package_candidate._records.md5_hash,
            package_candidate.size,
            base,
            destfile=destfile,
        )
        acq.run()

        if acqfile.status != acqfile.STAT_DONE:
            raise apt.package.FetchError(
                "The item %r could not be fetched: %s"
                % (acqfile.destfile, acqfile.error_text)
            )

        return os.path.abspath(destfile)


class Ubuntu(BaseRepo):
    @classmethod
    def get_package_libraries(cls, package_name):
        global _library_list
        if package_name not in _library_list:
            output = (
                subprocess.check_output(["dpkg", "-L", package_name])
                .decode(sys.getfilesystemencoding())
                .strip()
                .split()
            )
            _library_list[package_name] = {
                i for i in output if ("lib" in i and os.path.isfile(i))
            }

        return _library_list[package_name].copy()

    @classmethod
    def get_package_for_file(cls, file_path: str) -> str:
        return _run_dpkg_query_s(file_path)

    @classmethod
    def get_packages_for_source_type(cls, source_type):
        if source_type == "bzr":
            packages = {"bzr"}
        elif source_type == "git":
            packages = {"git"}
        elif source_type == "tar":
            packages = {"tar"}
        elif source_type == "hg" or source_type == "mercurial":
            packages = {"mercurial"}
        elif source_type == "subversion" or source_type == "svn":
            packages = {"subversion"}
        elif source_type == "rpm2cpio":
            packages = {"rpm2cpio"}
        elif source_type == "7zip":
            packages = {"p7zip-full"}
        else:
            packages = set()

        return packages

    @classmethod
    def refresh_build_packages(cls) -> None:
        try:
            subprocess.check_call(["sudo", "apt", "update"])
        except subprocess.CalledProcessError as call_error:
            raise errors.CacheUpdateFailedError(
                "failed to run apt update"
            ) from call_error

    @classmethod
    def install_build_packages(cls, package_names: List[str]) -> List[str]:
        """Install packages on the host required to build.

        :param package_names: a list of package names to install.
        :type package_names: a list of strings.
        :return: a list with the packages installed and their versions.
        :rtype: list of strings.
        :raises snapcraft.repo.errors.BuildPackageNotFoundError:
            if one of the packages was not found.
        :raises snapcraft.repo.errors.PackageBrokenError:
            if dependencies for one of the packages cannot be resolved.
        :raises snapcraft.repo.errors.BuildPackagesNotInstalledError:
            if installing the packages on the host failed.
        """
        new_packages = []  # type: List[Tuple[str, str]]
        with apt.Cache() as apt_cache:
            try:
                cls._mark_install(apt_cache, package_names)
            except errors.PackageNotFoundError as e:
                raise errors.BuildPackageNotFoundError(e.package_name)
            for package in apt_cache.get_changes():
                new_packages.append((package.name, package.candidate.version))

        if new_packages:
            cls._install_new_build_packages([package[0] for package in new_packages])
        return ["{}={}".format(package[0], package[1]) for package in new_packages]

    @classmethod
    def _mark_install(cls, apt_cache: apt.Cache, package_names: List[str]) -> None:
        for name in package_names:
            if name.endswith(":any"):
                name = name[:-4]
            if apt_cache.is_virtual_package(name):
                name = apt_cache.get_providing_packages(name)[0].name
            logger.debug(
                "Marking {!r} (and its dependencies) to be fetched".format(name)
            )
            name_arch, version = repo.get_pkg_name_parts(name)
            try:
                if version:
                    _set_pkg_version(apt_cache[name_arch], version)
                # Disable automatic resolving of broken packages here
                # because if that fails it raises a SystemError and the
                # API doesn't expose enough information about he problem.
                # Instead we let apt-get show a verbose error message later.
                # Also, make sure this package is marked as auto-installed,
                # which will propagate to its dependencies.
                apt_cache[name_arch].mark_install(auto_fix=False, from_user=False)

                # Now mark this package as NOT automatically installed, which
                # will leave its dependencies marked as auto-installed, which
                # allows us to clean them up if necessary.
                apt_cache[name_arch].mark_auto(False)

                cls._verify_marked_install(apt_cache[name_arch])
            except KeyError:
                raise errors.PackageNotFoundError(name)

    @classmethod
    def _verify_marked_install(cls, package: apt.Package):
        if not package.installed and not package.marked_install:
            broken_deps = []  # type: List[str]
            for deps in package.candidate.dependencies:
                for dep in deps:
                    if not dep.target_versions:
                        broken_deps.append(dep.name)
            raise errors.PackageBrokenError(package.name, broken_deps)

    @classmethod
    def _install_new_build_packages(cls, package_names: List[str]) -> None:
        package_names.sort()
        logger.info("Installing build dependencies: %s", " ".join(package_names))
        env = os.environ.copy()
        env.update(
            {
                "DEBIAN_FRONTEND": "noninteractive",
                "DEBCONF_NONINTERACTIVE_SEEN": "true",
                "DEBIAN_PRIORITY": "critical",
            }
        )

        apt_command = [
            "sudo",
            "--preserve-env",
            "apt-get",
            "--no-install-recommends",
            "-y",
        ]
        if not is_dumb_terminal():
            apt_command.extend(["-o", "Dpkg::Progress-Fancy=1"])
        apt_command.append("install")

        try:
            subprocess.check_call(apt_command + package_names, env=env)
        except subprocess.CalledProcessError:
            raise errors.BuildPackagesNotInstalledError(packages=package_names)

        try:
            subprocess.check_call(["sudo", "apt-mark", "auto"] + package_names, env=env)
        except subprocess.CalledProcessError as e:
            logger.warning(
                "Impossible to mark packages as auto-installed: {}".format(e)
            )

    @classmethod
    def build_package_is_valid(cls, package_name):
        with apt.Cache() as apt_cache:
            return package_name in apt_cache

    @classmethod
    def is_package_installed(cls, package_name):
        with apt.Cache() as apt_cache:
            if package_name not in apt_cache:
                return False
            return apt_cache[package_name].installed

    @classmethod
    def get_installed_packages(cls):
        installed_packages = []
        with apt.Cache() as apt_cache:
            for package in apt_cache:
                if package.installed:
                    installed_packages.append(
                        "{}={}".format(package.name, package.installed.version)
                    )
        return installed_packages

    def __init__(
        self, rootdir, sources=None, keyrings=None, project_options=None
    ) -> None:
        super().__init__(rootdir)
        self._downloaddir = os.path.join(rootdir, "download")

        if not project_options:
            project_options = snapcraft.ProjectOptions()

        self._apt = _AptCache(
            project_options.deb_arch, sources_list=sources, keyrings=keyrings
        )

        self._cache = cache.AptStagePackageCache(
            sources_digest=self._apt.sources_digest()
        )

    def is_valid(self, package_name):
        with self._apt.archive(self._cache.base_dir) as apt_cache:
            return package_name in apt_cache

    def get(self, package_names) -> None:
        with self._apt.archive(self._cache.base_dir) as apt_cache:
            self._mark_install(apt_cache, package_names)
            self._filter_base_packages(apt_cache, package_names)
            self._autokeep_packages(apt_cache)
            return self._get(apt_cache)

    def _autokeep_packages(self, apt_cache):
        for package in apt_cache.get_changes():
            if package.is_auto_removable:
                package.mark_keep()

    def _filter_base_packages(self, apt_cache, package_names):
        manifest_dep_names = self._manifest_dep_names(apt_cache)

        skipped_essential = []
        skipped_blacklisted = []

        # unmark some base packages here
        # note that this will break the consistency check inside apt_cache
        # (apt_cache.broken_count will be > 0)
        # but that is ok as it was consistent before we excluded
        # these base package
        for pkg in apt_cache:
            # those should be already on each system, it also prevents
            # diving into downloading libc6
            if pkg.candidate.priority in "essential" and pkg.name not in package_names:
                skipped_essential.append(pkg.name)
                pkg.mark_keep()
                continue
            if pkg.name in manifest_dep_names and pkg.name not in package_names:
                skipped_blacklisted.append(pkg.name)
                pkg.mark_keep()
                continue

        if skipped_essential:
            logger.debug(
                "Skipping priority essential packages: "
                "{!r}".format(skipped_essential)
            )
        if skipped_blacklisted:
            logger.debug(
                "Skipping blacklisted from manifest packages: "
                "{!r}".format(skipped_blacklisted)
            )

    def _get(self, apt_cache):
        # Ideally we'd use apt.Cache().fetch_archives() here, but it seems to
        # mangle some package names on disk such that we can't match it up to
        # the archive later. We could get around this a few different ways:
        #
        # 1. Store each stage package in the cache named by a hash instead of
        #    its name from the archive.
        # 2. Download packages in a different manner.
        #
        # In the end, (2) was chosen for minimal overhead and a simpler cache
        # implementation. So we're using fetch_binary() here instead.
        # Downloading each package individually has the drawback of witholding
        # any clue of how long the whole pulling process will take, but that's
        # something we'll have to live with.
        pkg_list = []
        for package in apt_cache.get_changes():
            pkg_list.append(str(package.candidate))
            try:
                source = self._apt.fetch_binary(
                    package_candidate=package.candidate,
                    destination=self._cache.packages_dir,
                )
            except apt.package.FetchError as e:
                raise errors.PackageFetchError(str(e))
            destination = os.path.join(self._downloaddir, os.path.basename(source))
            with contextlib.suppress(FileNotFoundError):
                os.remove(destination)
            file_utils.link_or_copy(source, destination)

        return pkg_list

    def unpack(self, unpackdir) -> None:
        pkgs_abs_path = glob.glob(os.path.join(self._downloaddir, "*.deb"))
        for pkg in pkgs_abs_path:
            # TODO needs elegance and error control
            try:
                subprocess.check_call(["dpkg-deb", "--extract", pkg, unpackdir])
            except subprocess.CalledProcessError:
                raise errors.UnpackError(pkg)
        self.normalize(unpackdir)

    def _manifest_dep_names(self, apt_cache):
        manifest_dep_names = set()

        with open(os.path.abspath(os.path.join(__file__, "..", "manifest.txt"))) as f:
            for line in f:
                pkg = line.strip()
                if pkg in apt_cache:
                    manifest_dep_names.add(pkg)

        return manifest_dep_names


def _get_local_sources_list():
    sources_list = glob.glob("/etc/apt/sources.list.d/*.list")
    sources_list.append("/etc/apt/sources.list")

    sources = ""
    for source in sources_list:
        with open(source) as f:
            sources += f.read()

    return sources


def _get_geoip_country_code_prefix():
    try:
        with urllib.request.urlopen(_GEOIP_SERVER) as f:
            xml_data = f.read()
        et = ElementTree.fromstring(xml_data)
        cc = et.find("CountryCode")
        if cc is None:
            return ""
        return cc.text.lower()
    except (ElementTree.ParseError, urllib.error.URLError):
        pass
    return ""


def _format_sources_list(sources_list, *, deb_arch, use_geoip=False, release="xenial"):
    if not sources_list:
        sources_list = _DEFAULT_SOURCES

    if deb_arch in ("amd64", "i386"):
        if use_geoip:
            geoip_prefix = _get_geoip_country_code_prefix()
            prefix = "{}.archive".format(geoip_prefix)
        else:
            prefix = "archive"
        suffix = "ubuntu"
        security = "security"
    else:
        prefix = "ports"
        suffix = "ubuntu-ports"
        security = "ports"

    return string.Template(sources_list).substitute(
        {"prefix": prefix, "release": release, "suffix": suffix, "security": security}
    )


def _fix_filemode(path):
    mode = stat.S_IMODE(os.stat(path, follow_symlinks=False).st_mode)
    if mode & 0o4000 or mode & 0o2000:
        logger.warning("Removing suid/guid from {}".format(path))
        os.chmod(path, mode & 0o1777)


def _try_copy_local(path, target):
    real_path = os.path.realpath(path)
    if os.path.exists(real_path):
        logger.warning(
            "Copying needed target link from the system {}".format(real_path)
        )
        os.makedirs(os.path.dirname(target), exist_ok=True)
        shutil.copyfile(os.readlink(path), target)
        return True
    else:
        logger.warning("{} will be a dangling symlink".format(path))
        return False


def _set_pkg_version(pkg, version):
    """Set cadidate version to a specific version if available"""
    if version in pkg.versions:
        version = pkg.versions.get(version)
        pkg.candidate = version
    else:
        raise errors.PackageNotFoundError("{}={}".format(pkg.name, version))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
import json
import logging
import os
import re
import shutil
import stat
import subprocess
import sys
import tempfile
from typing import List, Optional, Sequence, Set

import snapcraft
from snapcraft import file_utils
from snapcraft.internal import mangling
from ._python_finder import get_python_command, get_python_headers, get_python_home
from . import errors

logger = logging.getLogger(__name__)


def _process_common_args(
    *, constraints: Optional[Set[str]] = None, process_dependency_links: bool = False
) -> List[str]:
    args = []
    if constraints:
        for constraint in constraints:
            args.extend(["--constraint", constraint])

    if process_dependency_links:
        args.append("--process-dependency-links")

    return args


def _process_package_args(
    *, packages: Sequence[str], requirements: Sequence[str], setup_py_dir: str
) -> List[str]:
    args = []
    if requirements:
        for requirement in requirements:
            args.extend(["--requirement", requirement])

    if packages:
        args.extend(packages)

    if setup_py_dir:
        args.append(".")

    return args


def _replicate_owner_mode(path):
    # Don't bother with a path that doesn't exist or is a symlink. The target
    # of the symlink will either be updated anyway, or we won't have permission
    # to change it.
    if not os.path.exists(path) or os.path.islink(path):
        return

    file_mode = os.stat(path).st_mode

    # We at least need to write to it to fix shebangs later
    new_mode = file_mode | stat.S_IWUSR

    # If the owner can execute it, so should everyone else.
    if file_mode & stat.S_IXUSR:
        new_mode |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH

    # If the owner can read it, so should everyone else
    if file_mode & stat.S_IRUSR:
        new_mode |= stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

    os.chmod(path, new_mode)


def _fix_permissions(path):
    for root, dirs, files in os.walk(path):
        for filename in files:
            _replicate_owner_mode(os.path.join(root, filename))
        for dirname in dirs:
            _replicate_owner_mode(os.path.join(root, dirname))


class Pip:
    """Wrapper for pip abstracting the args necessary for use in a part.

    This class takes care of fetching pip, setuptools, and wheel, and then
    simply shells out to pip with the magical arguments necessary to install
    packages into a part.

    Of particular importance: packages must be downloaded (via download())
    before they can be installed or have wheels built.
    """

    def __init__(self, *, python_major_version, part_dir, install_dir, stage_dir):
        """Initialize pip.

        You must call setup() before you can actually use pip.

        :param str python_major_version: The python major version to find (2 or
                                         3)
        :param str part_dir: Path to the part's working area
        :param str install_dir: Path to the part's install area
        :param str stage_dir: Path to the staging area

        :raises MissingPythonCommandError: If no python could be found in the
                                           staging or part's install area.
        """
        self._python_major_version = python_major_version
        self._install_dir = install_dir
        self._stage_dir = stage_dir

        self._python_package_dir = os.path.join(part_dir, "python-packages")
        os.makedirs(self._python_package_dir, exist_ok=True)

        self.__python_command = None  # type:str
        self.__python_home = None  # type: str

    @property
    def _python_command(self):
        """Lazily determine the python command required."""
        if not self.__python_command:
            self.__python_command = get_python_command(
                self._python_major_version,
                stage_dir=self._stage_dir,
                install_dir=self._install_dir,
            )
        return self.__python_command

    @property
    def _python_home(self):
        """Lazily determine the correct python home."""
        if not self.__python_home:
            self.__python_home = get_python_home(
                self._python_major_version,
                stage_dir=self._stage_dir,
                install_dir=self._install_dir,
            )
        return self.__python_home

    def setup(self):
        """Install pip and dependencies.

        Check to see if pip has already been installed. If not, fetch pip,
        setuptools, and wheel, and install them so they can be used.
        """

        self._ensure_pip_installed()
        self._ensure_wheel_installed()
        self._ensure_setuptools_installed()

    def is_setup(self):
        """Return true if this class has already been setup."""

        return (
            self._is_pip_installed()
            and self._is_wheel_installed()
            and self._is_setuptools_installed()
        )

    def _ensure_pip_installed(self):
        # Check to see if we have our own pip. If not, we need to use the pip
        # on the host (installed via build-packages) to grab our own.
        if not self._is_pip_installed():
            logger.info("Fetching and installing pip...")

            real_python_home = self.__python_home

            # Make sure we're using pip from the host. Wrapping this operation
            # in a try/finally to make sure we revert away from the host's
            # python at the end.
            try:
                self.__python_home = os.path.join(os.path.sep, "usr")

                # Using the host's pip, install our own pip
                self.download({"pip"})
                self.install({"pip"}, ignore_installed=True)
            finally:
                # Now that we have our own pip, reset the python home
                self.__python_home = real_python_home

    def _ensure_wheel_installed(self):
        if not self._is_wheel_installed():
            logger.info("Fetching and installing wheel...")
            self.download({"wheel"})
            self.install({"wheel"}, ignore_installed=True)

    def _ensure_setuptools_installed(self):
        if not self._is_setuptools_installed():
            logger.info("Fetching and installing setuptools...")
            self.download({"setuptools"})
            self.install({"setuptools"}, ignore_installed=True)

    def _is_pip_installed(self):
        try:
            # We're expecting an error here at least once complaining about
            # pip not being installed. In order to verify that the error is the
            # one we think it is, we need to process the stderr. So we'll
            # redirect it to stdout. If it's not the error we expect, something
            # is wrong, so re-raise it.
            #
            # Using _run_output here so stdout doesn't get printed to the
            # terminal.
            self._run_output([], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            output = e.output.decode(sys.getfilesystemencoding()).strip()
            if "no module named pip" in output.lower():
                return False
            else:
                raise e
        return True

    def _is_wheel_installed(self):
        return "wheel" in self.list()

    def _is_setuptools_installed(self):
        return "setuptools" in self.list()

    def download(
        self,
        packages,
        *,
        setup_py_dir: Optional[str] = None,
        constraints: Optional[Set[str]] = None,
        requirements: Optional[Sequence[str]] = None,
        process_dependency_links: bool = False
    ):
        """Download packages into cache, but do not install them.

        :param iterable packages: Packages to download from index.
        :param str setup_py_dir: Directory containing setup.py.
        :param iterable constraints: Collection of paths to constraints files.
        :param iterable requirements: Collection of paths to requirements
                                      files.
        :param boolean process_dependency_links: Enable the processing of
                                                 dependency links.
        """
        package_args = _process_package_args(
            packages=packages, requirements=requirements, setup_py_dir=setup_py_dir
        )

        if not package_args:
            return  # No operation was requested

        args = _process_common_args(
            process_dependency_links=process_dependency_links, constraints=constraints
        )

        # Using pip with a few special parameters:
        #
        # --disable-pip-version-check: Don't whine if pip is out-of-date with
        #                              the version on pypi.
        # --dest: Download packages into the directory we've set aside for it.
        #
        # For cwd, setup_py_dir will be the actual directory we need to be in
        # or None.
        self._run(
            [
                "download",
                "--disable-pip-version-check",
                "--dest",
                self._python_package_dir,
            ]
            + args
            + package_args,
            cwd=setup_py_dir,
        )

    def install(
        self,
        packages,
        *,
        setup_py_dir: Optional[str] = None,
        constraints: Optional[Set[str]] = None,
        requirements: Optional[Sequence[str]] = None,
        process_dependency_links: bool = False,
        upgrade: bool = False,
        install_deps: bool = True,
        ignore_installed: bool = False
    ):
        """Install packages from cache.

        The packages should have already been downloaded via `download()`.

        :param iterable packages: Packages to install from cache.
        :param str setup_py_dir: Directory containing setup.py.
        :param iterable constraints: Collection of paths to constraints files.
        :param iterable requirements: Collection of paths to requirements
                                      files.
        :param boolean process_dependency_links: Enable the processing of
                                                 dependency links.
        :param boolean upgrade: Recursively upgrade packages.
        :param boolean install_deps: Install package dependencies.
        :param boolean ignore_installed: Reinstall packages if they're already
                                         installed
        """
        package_args = _process_package_args(
            packages=packages, requirements=requirements, setup_py_dir=setup_py_dir
        )

        if not package_args:
            return  # No operation was requested

        args = _process_common_args(
            process_dependency_links=process_dependency_links, constraints=constraints
        )

        if upgrade:
            args.append("--upgrade")

        if not install_deps:
            args.append("--no-deps")

        if ignore_installed:
            args.append("--ignore-installed")

        # Using pip with a few special parameters:
        #
        # --user: Install packages to PYTHONUSERBASE, which we've pointed to
        #         the installdir.
        # --no-compile: Don't compile .pyc files. FIXME: This is legacy, and
        #               should be removed once this refactor has been
        #               validated.
        # --no-index: Don't hit pypi, assume the packages are already
        #             downloaded (i.e. by using `self.download()`)
        # --find-links: Provide the directory into which the packages should
        #               have already been fetched
        #
        # For cwd, setup_py_dir will be the actual directory we need to be in
        # or None.
        self._run(
            [
                "install",
                "--user",
                "--no-compile",
                "--no-index",
                "--find-links",
                self._python_package_dir,
            ]
            + args
            + package_args,
            cwd=setup_py_dir,
        )

        # Installing with --user results in a directory with 700 permissions.
        # We need it a bit more open than that, so open it up.
        _fix_permissions(self._install_dir)

        # Fix all shebangs to use the in-snap python.
        mangling.rewrite_python_shebangs(self._install_dir)

    def wheel(
        self,
        packages,
        *,
        setup_py_dir: Optional[str] = None,
        constraints: Optional[Set[str]] = None,
        requirements: Optional[Sequence[str]] = None,
        process_dependency_links: bool = False
    ):
        """Build wheels of packages in the cache.

        The packages should have already been downloaded via `download()`.

        :param iterable packages: Packages in cache for which to build wheels.
        :param str setup_py_dir: Directory containing setup.py.
        :param iterable constraints: Collection of paths to constraints files.
        :param iterable requirements: Collection of paths to requirements
                                      files.
        :param boolean process_dependency_links: Enable the processing of
                                                 dependency links.

        :return: List of paths to each wheel that was built.
        :rtype: list
        """
        package_args = _process_package_args(
            packages=packages, requirements=requirements, setup_py_dir=setup_py_dir
        )

        if not package_args:
            return []  # No operation was requested

        args = _process_common_args(
            process_dependency_links=process_dependency_links, constraints=constraints
        )

        wheels = []  # type: List[str]
        with tempfile.TemporaryDirectory() as temp_dir:

            # Using pip with a few special parameters:
            #
            # --no-index: Don't hit pypi, assume the packages are already
            #             downloaded (i.e. by using `self.download()`)
            # --find-links: Provide the directory into which the packages
            #               should have already been fetched
            # --wheel-dir: Build wheels into a temporary working area rather
            #              rather than cwd. We'll copy them over. FIXME: We can
            #              probably get away just building them in the package
            #              dir. Try that once this refactor has been validated.
            self._run(
                [
                    "wheel",
                    "--no-index",
                    "--find-links",
                    self._python_package_dir,
                    "--wheel-dir",
                    temp_dir,
                ]
                + args
                + package_args,
                cwd=setup_py_dir,
            )
            wheels = os.listdir(temp_dir)
            for wheel in wheels:
                file_utils.link_or_copy(
                    os.path.join(temp_dir, wheel),
                    os.path.join(self._python_package_dir, wheel),
                )

        return [os.path.join(self._python_package_dir, wheel) for wheel in wheels]

    def list(self, *, user=False):
        """Determine which packages have been installed.

        :param boolean user: Whether or not to limit results to user base.

        :return: Dict of installed python packages and their versions
        :rtype: dict
        """
        command = ["list"]
        if user:
            command.append("--user")

        packages = collections.OrderedDict()
        try:
            output = self._run_output(command + ["--format=json"])
            json_output = json.loads(output, object_pairs_hook=collections.OrderedDict)
        except subprocess.CalledProcessError:
            # --format requires a newer pip, so fall back to legacy output
            output = self._run_output(command)
            json_output = []  # type: List[Dict[str, str]]
            version_regex = re.compile(r"\((.+)\)")
            for line in output.splitlines():
                line = line.split()
                m = version_regex.search(line[1])
                if not m:
                    raise errors.PipListInvalidLegacyFormatError(output)
                json_output.append({"name": line[0], "version": m.group(1)})
        except json.decoder.JSONDecodeError as e:
            raise errors.PipListInvalidJsonError(output) from e

        for package in json_output:
            if "name" not in package:
                raise errors.PipListMissingFieldError("name", output)
            if "version" not in package:
                raise errors.PipListMissingFieldError("version", output)
            packages[package["name"]] = package["version"]
        return packages

    def clean_packages(self):
        """Remove the package cache."""
        with contextlib.suppress(FileNotFoundError):
            shutil.rmtree(self._python_package_dir)

    def env(self):
        """The environment used by pip.

        This function is only useful if you happen to need to call into pip's
        environment without using the API otherwise made available here (e.g.
        calling the setup.py directly instead of with pip).

        :return: Dict of the environment necessary to use the pip contained
                 here.
        :rtype: dict
        """
        env = os.environ.copy()
        env["PYTHONUSERBASE"] = self._install_dir
        env["PYTHONHOME"] = self._python_home

        env["PATH"] = "{}:{}".format(
            os.path.join(self._install_dir, "usr", "bin"), os.path.expandvars("$PATH")
        )

        headers = get_python_headers(
            self._python_major_version, stage_dir=self._stage_dir
        )
        if headers:
            current_cppflags = env.get("CPPFLAGS", "")
            env["CPPFLAGS"] = "-I{}".format(headers)
            if current_cppflags:
                env["CPPFLAGS"] = "{} {}".format(env["CPPFLAGS"], current_cppflags)

        return env

    def _run(self, args, runner=None, **kwargs):
        env = self.env()

        # Using None as the default value instead of common.run so we can mock
        # common.run.
        if runner is None:
            runner = snapcraft.internal.common.run

        return runner(
            [self._python_command, "-m", "pip"] + list(args), env=env, **kwargs
        )

    def _run_output(self, args, **kwargs):
        return self._run(args, runner=snapcraft.internal.common.run_output, **kwargs)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pkgutil

import click

import snapcraft
from snapcraft.internal.common import format_output_in_columns
from snapcraft.internal.common import get_terminal_width


@click.group()
def discoverycli():
    pass


@discoverycli.command("list-plugins")
def list_plugins():
    """List the available plugins that handle different types of part.

    This command has an alias of `plugins`.
    """
    plugins = []
    for _, modname, _ in pkgutil.iter_modules(snapcraft.plugins.__path__):
        # Only add non-private modules/packages to the plugin list
        if not modname.startswith("_"):
            plugins.append(modname.replace("_", "-"))

    # we wrap the output depending on terminal size
    width = get_terminal_width()
    for line in format_output_in_columns(plugins, max_width=width):
        click.echo(line)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import importlib
import logging
import os
import sys

import jsonschema

import snapcraft
from snapcraft.project.errors import YamlValidationError
from snapcraft.internal import errors
from typing import Optional

logger = logging.getLogger(__name__)


def load_plugin(
    plugin_name,
    part_name,
    project_options,
    properties,
    part_schema,
    definitions_schema,
    local_plugins_dir: Optional[str] = None,
):
    if not local_plugins_dir:
        local_plugins_dir = os.path.join(os.getcwd(), "snap", "plugins")

    module_name = plugin_name.replace("-", "_")
    module = _load_module(module_name, plugin_name, local_plugins_dir)
    plugin_class = _get_plugin(module)
    if not plugin_class:
        raise errors.PluginError("no plugin found in module {!r}".format(plugin_name))

    _validate_pull_and_build_properties(
        plugin_name, plugin_class, part_schema, definitions_schema
    )

    try:
        options = _make_options(
            part_schema, definitions_schema, properties, plugin_class.schema()
        )
    except jsonschema.ValidationError as e:
        error = YamlValidationError.from_validation_error(e)
        raise errors.PluginError(
            "properties failed to load for {}: {}".format(part_name, error.message)
        )

    plugin = plugin_class(part_name, options, project_options)

    if project_options.is_cross_compiling:
        logger.debug(
            "Setting {!r} as the compilation target for {!r}".format(
                project_options.deb_arch, plugin_name
            )
        )
        plugin.enable_cross_compilation()

    return plugin


def _load_module(module_name, plugin_name, local_plugins_dir):
    module = None
    with contextlib.suppress(ImportError):
        module = _load_local("x-{}".format(plugin_name), local_plugins_dir)
        logger.info("Loaded local plugin for %s", plugin_name)

    if not module:
        with contextlib.suppress(ImportError):
            module = importlib.import_module("snapcraft.plugins.{}".format(module_name))

    if not module:
        logger.info("Searching for local plugin for %s", plugin_name)
        with contextlib.suppress(ImportError):
            module = _load_local(module_name, local_plugins_dir)
        if not module:
            raise errors.PluginError("unknown plugin: {!r}".format(plugin_name))

    return module


def _load_local(module_name, local_plugin_dir):
    sys.path = [local_plugin_dir] + sys.path
    try:
        module = importlib.import_module(module_name)
    finally:
        sys.path.pop(0)

    return module


def _get_plugin(module):
    for attr in vars(module).values():
        if not isinstance(attr, type):
            continue
        if not issubclass(attr, snapcraft.BasePlugin):
            continue
        if attr == snapcraft.BasePlugin:
            continue
        return attr


def _validate_pull_and_build_properties(
    plugin_name, plugin, part_schema, definitions_schema
):
    merged_schema = _merged_part_and_plugin_schemas(
        part_schema, definitions_schema, plugin.schema()
    )
    merged_properties = merged_schema["properties"]

    # First, validate pull properties
    invalid_properties = _validate_step_properties(
        plugin.get_pull_properties(), merged_properties
    )

    if invalid_properties:
        raise errors.InvalidPullPropertiesError(plugin_name, list(invalid_properties))

    # Now, validate build properties
    invalid_properties = _validate_step_properties(
        plugin.get_build_properties(), merged_properties
    )

    if invalid_properties:
        raise errors.InvalidBuildPropertiesError(plugin_name, list(invalid_properties))


def _validate_step_properties(step_properties, schema_properties):
    invalid_properties = set()
    for step_property in step_properties:
        if step_property not in schema_properties:
            invalid_properties.add(step_property)

    return invalid_properties


def _make_options(part_schema, definitions_schema, properties, plugin_schema):
    # Make copies as these dictionaries are tampered with
    part_schema = part_schema.copy()
    properties = properties.copy()

    plugin_schema = _merged_part_and_plugin_schemas(
        part_schema, definitions_schema, plugin_schema
    )

    jsonschema.validate(properties, plugin_schema)

    options = _populate_options(properties, plugin_schema)

    return options


def _merged_part_and_plugin_schemas(part_schema, definitions_schema, plugin_schema):
    plugin_schema = plugin_schema.copy()
    if "properties" not in plugin_schema:
        plugin_schema["properties"] = {}

    if "definitions" not in plugin_schema:
        plugin_schema["definitions"] = {}

    # The part schema takes precedence over the plugin's schema.
    plugin_schema["properties"].update(part_schema)
    plugin_schema["definitions"].update(definitions_schema)

    return plugin_schema


def _populate_options(properties, schema):
    class Options:
        pass

    options = Options()

    schema_properties = schema.get("properties", {})
    for key in schema_properties:
        attr_name = key.replace("-", "_")
        default_value = schema_properties[key].get("default")
        attr_value = properties.get(key, default_value)
        setattr(options, attr_name, attr_value)

    return options



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Marius Gripsgard (mariogrip@ubuntu.com)
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This rust plugin is useful for building rust based parts.

Rust uses cargo to drive the build.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - rust-channel
      (string)
      select rust channel (stable, beta, nightly)
    - rust-revision
      (string)
      select rust version, this is preferred over rust-channel
    - rust-features
      (list of strings)
      Features used to build optional dependencies

If a rust-toolchain file is found, the toolchain specified there will
be used unless rust-channel or rust-revision are set which will
override the rust-toolchain file. If neither rust-toolchain exists nor
rust-channel or rust-revision are set, the latest stable toolchain
will be used.
"""

import collections
import logging
import os
from contextlib import suppress
from typing import List, Optional

import toml

import snapcraft
from snapcraft import sources
from snapcraft import shell_utils
from snapcraft.internal import errors

_RUSTUP = "https://sh.rustup.rs/"
logger = logging.getLogger(__name__)


class RustPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["rust-channel"] = {
            "type": "string",
            "enum": ["stable", "beta", "nightly"],
        }
        schema["properties"]["rust-revision"] = {"type": "string"}
        schema["properties"]["rust-features"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["rust-revision", "rust-channel"]

    @classmethod
    def get_build_properties(cls):
        return ["rust-features"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        if project.info.get_build_base() not in ("core", "core16", "core18"):
            raise errors.PluginBaseError(
                part_name=self.name, base=project.info.get_build_base()
            )

        self.build_packages.extend(["gcc", "git", "curl", "file"])
        self._rust_dir = os.path.expanduser(os.path.join("~", ".cargo"))
        self._rustup_cmd = os.path.join(self._rust_dir, "bin", "rustup")
        self._cargo_cmd = os.path.join(self._rust_dir, "bin", "cargo")
        self._rustc_cmd = os.path.join(self._rust_dir, "bin", "rustc")
        self._rustdoc_cmd = os.path.join(self._rust_dir, "bin", "rustdoc")

        self._manifest = collections.OrderedDict()

    def enable_cross_compilation(self):
        # The logic is applied transparently trough internal
        # rust tooling.
        pass

    def pull(self):
        super().pull()
        self._fetch_rustup()
        self._fetch_rust()
        self._fetch_cargo_deps()

    def _fetch_rustup(self):
        # if rustup-init has already been done, we can skip this.
        if os.path.exists(os.path.join(self._rust_dir, "bin", "rustup")):
            return

        # Download rustup-init.
        os.makedirs(self._rust_dir, exist_ok=True)
        rustup_init_cmd = os.path.join(self._rust_dir, "rustup.sh")
        sources.Script(_RUSTUP, self._rust_dir).download(filepath=rustup_init_cmd)

        # Basic options:
        # -y: assume yes
        # --no-modify-path: do not modify bashrc
        options = ["-y", "--no-modify-path"]
        if self._get_toolchain() is not None:
            options.extend(["--default-toolchain", "none"])

        # Fetch rust
        self.run([rustup_init_cmd] + options, env=self._build_env())

    def _fetch_rust(self) -> None:
        toolchain = self._get_toolchain()

        # We only do this when one of the snapcraft properties are set,
        # the reason for this is faster iterations when changing any
        # of the toolchain related properties.
        # When changing the rust-toolchain file a re-pull will be triggered
        # due to source changes and the right thing will happen.
        if toolchain is not None:
            # https://rust-lang-nursery.github.io/edition-guide/rust-2018/rustup-for-managing-rust-versions.html
            self.run([self._rustup_cmd, "install", toolchain], env=self._build_env())

        # Add the appropriate target cross compilation target if necessary.
        # https://github.com/rust-lang/rustup.rs/blob/master/README.md#cross-compilation
        if self.project.is_cross_compiling:
            add_target_cmd = [self._rustup_cmd, "target", "add"]
            if toolchain is not None:
                add_target_cmd.extend(["--toolchain", toolchain])
            add_target_cmd.append(self._get_target())

            self.run(add_target_cmd, env=self._build_env())

    def _fetch_cargo_deps(self):
        if self.options.source_subdir:
            sourcedir = os.path.join(self.sourcedir, self.options.source_subdir)
        else:
            sourcedir = self.sourcedir

        fetch_cmd = [
            self._cargo_cmd,
            "fetch",
            "--manifest-path",
            os.path.join(sourcedir, "Cargo.toml"),
        ]
        toolchain = self._get_toolchain()
        if toolchain is not None:
            fetch_cmd.insert(1, "+{}".format(toolchain))
        self.run(fetch_cmd, env=self._build_env())

    def _get_target(self) -> str:
        # Cf. rustc --print target-list
        targets = {
            "armhf": "armv7-{}-{}eabihf",
            "arm64": "aarch64-{}-{}",
            "i386": "i686-{}-{}",
            "amd64": "x86_64-{}-{}",
            "ppc64el": "powerpc64le-{}-{}",
            "s390x": "s390x-{}-{}",
        }
        rust_target = targets.get(self.project.deb_arch)
        if not rust_target:
            raise errors.SnapcraftEnvironmentError(
                "{!r} is not supported as a target architecture ".format(
                    self.project.deb_arch
                )
            )
        return rust_target.format("unknown-linux", "gnu")

    def build(self):
        super().build()

        # Write a minimal config.
        self._write_cargo_config()

        install_cmd = [
            self._cargo_cmd,
            "install",
            "--path",
            self.builddir,
            "--root",
            self.installdir,
            "--force",
        ]
        toolchain = self._get_toolchain()
        if toolchain is not None:
            install_cmd.insert(1, "+{}".format(toolchain))

        # Even though this is mostly harmless when not cross compiling
        # the flag is in place to avoid a situation where an earlier
        # version of the toolchain is used.
        if self.project.is_cross_compiling:
            install_cmd.extend(["--target", self._get_target()])

        if self.options.rust_features:
            install_cmd.append("--features")
            install_cmd.append(" ".join(self.options.rust_features))

        # build and install.
        self.run(install_cmd, env=self._build_env())

        # Finally, record.
        self._record_manifest()

    def _build_env(self):
        env = os.environ.copy()

        env.update(dict(RUSTUP_HOME=self._rust_dir, CARGO_HOME=self._rust_dir))

        rustflags = self._get_rustflags()
        if rustflags:
            string_fmt = " ".join(["{}".format(i) for i in rustflags]).strip()
            env.update(RUSTFLAGS=string_fmt)

        return env

    def _get_toolchain(self) -> Optional[str]:
        """Return the toolchain to use.

        If a rust-toolchain file is present, None will be returned.
        If a rust-toolchain file is not present and neither rust-version
        nor rust-channel are set, then stable will be returned.
        """
        if self.options.source_subdir:
            sourcedir = os.path.join(self.sourcedir, self.options.source_subdir)
        else:
            sourcedir = self.sourcedir
        rust_toolchain_path = os.path.join(sourcedir, "rust-toolchain")

        if self.options.rust_revision:
            toolchain = self.options.rust_revision
        elif self.options.rust_channel:
            toolchain = self.options.rust_channel
        elif not os.path.exists(rust_toolchain_path):
            toolchain = "stable"
        else:
            toolchain = None

        return toolchain

    def _get_linker(self) -> str:
        arch_triplet = self.project.arch_triplet
        if arch_triplet == "i386-linux-gnu":
            return "i686-linux-gnu-gcc"
        else:
            return "{}-gcc".format(arch_triplet)

    def _write_cargo_config(self, cargo_config_path: Optional[str] = None) -> None:
        if cargo_config_path is None:
            cargo_config_path = os.path.join(self.builddir, ".cargo", "config")
        if os.path.dirname(cargo_config_path):
            os.makedirs(os.path.dirname(cargo_config_path), exist_ok=True)

        target = {self._get_target(): dict(linker=self._get_linker())}
        config = dict(
            arch_triplet=self.project.arch_triplet,
            jobs=self.parallel_build_count,
            rustc_cmd=self._rustc_cmd,
            rustdoc_cmd=self._rustdoc_cmd,
            target=target,
        )

        # Cf. http://doc.crates.io/config.html
        with open(cargo_config_path, "w") as toml_config_file:
            toml.dump(config, toml_config_file)

    def _get_rustflags(self) -> List[str]:
        ldflags = shell_utils.getenv("LDFLAGS")
        rustldflags = []
        flags = {flag for flag in ldflags.split(" ") if flag}
        for flag in flags:
            rustldflags.extend(["-C", "link-arg={}".format(flag)])

        if self.project.is_cross_compiling:
            rustldflags.extend(["-C", "linker={}".format(self._get_linker())])

        return rustldflags

    def _record_manifest(self):
        self._manifest["rustup-version"] = self.run_output(
            [self._rustup_cmd, "--version"], env=self._build_env()
        )
        toolchain = self._get_toolchain()
        if toolchain is not None:
            toolchain_option = "+{}".format(toolchain)
            self._manifest["rustc-version"] = self.run_output(
                [self._rustc_cmd, toolchain_option, "--version"], env=self._build_env()
            )
            self._manifest["cargo-version"] = self.run_output(
                [self._cargo_cmd, toolchain_option, "--version"], env=self._build_env()
            )
        else:
            self._manifest["rustc-version"] = self.run_output(
                [self._rustc_cmd, "--version"], env=self._build_env()
            )
            self._manifest["cargo-version"] = self.run_output(
                [self._cargo_cmd, "--version"], env=self._build_env()
            )
        with suppress(FileNotFoundError, IsADirectoryError):
            with open(os.path.join(self.builddir, "Cargo.lock")) as lock_file:
                self._manifest["cargo-lock-contents"] = lock_file.read()

    def get_manifest(self):
        return self._manifest



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, RetryError
from requests.packages.urllib3.util.retry import Retry
import urllib.parse
import os

from . import _agent
from . import errors

# Set urllib3's logger to only emit errors, not warnings. Otherwise even
# retries are printed, and they're nasty.
logging.getLogger(requests.packages.urllib3.__package__).setLevel(logging.ERROR)
logger = logging.getLogger(__name__)


class Client:
    """A base class to define clients for the ols servers.
    This is a simple wrapper around requests.Session so we inherit all good
    bits while providing a simple point for tests to override when needed.
    """

    def __init__(self, conf, root_url):
        """Initialize Client object

        :param config conf: Configuration details for the client
        :param str root_url: Root url for all requests.
        :type config: snapcraft.config.Config
        """
        self.conf = conf
        self.root_url = root_url
        self.session = requests.Session()
        # Setup max retries for all store URLs and the CDN
        retries = Retry(
            total=int(os.environ.get("STORE_RETRIES", 5)),
            backoff_factor=int(os.environ.get("STORE_BACKOFF", 2)),
            status_forcelist=[104, 500, 502, 503, 504],
        )
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        self._snapcraft_headers = {"User-Agent": _agent.get_user_agent()}

    def request(self, method, url, params=None, headers=None, **kwargs):
        """Send a request to url relative to the root url.

        :param str method: Method used for the request.
        :param str url: Appended with the root url first.
        :param list params: Query parameters to be sent along with the request.
        :param list headers: Headers to be sent along with the request.

        :return Response of the request.
        """
        # Note that url may be absolute in which case 'root_url' is ignored by
        # urljoin.

        if headers:
            headers.update(self._snapcraft_headers)
        else:
            headers = self._snapcraft_headers

        final_url = urllib.parse.urljoin(self.root_url, url)
        debug_headers = headers.copy()
        if "Authorization" in debug_headers:
            debug_headers["Authorization"] = "<macaroon>"
        logger.debug(
            "Calling {} with params {} and headers {}".format(
                final_url, params, debug_headers
            )
        )
        try:
            response = self.session.request(
                method, final_url, headers=headers, params=params, **kwargs
            )
        except (ConnectionError, RetryError) as e:
            raise errors.StoreNetworkError(e) from e

        # Handle 5XX responses generically right here, so the callers don't
        # need to worry about it.
        if response.status_code >= 500:
            raise errors.StoreServerError(response)

        return response

    def get(self, url, **kwargs):
        """Perform a GET request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self.request("GET", url, **kwargs)

    def post(self, url, **kwargs):
        """Perform a POST request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self.request("POST", url, **kwargs)

    def put(self, url, **kwargs):
        """Perform a PUT request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self.request("PUT", url, **kwargs)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import shutil
import subprocess
import tempfile

from ._cache import SnapcraftProjectCache
from snapcraft import file_utils, yaml_utils

logger = logging.getLogger(__name__)


class SnapCache(SnapcraftProjectCache):
    """Cache for snap revisions."""

    def __init__(self, *, project_name):
        super().__init__(project_name=project_name)
        self.snap_cache_root = self._setup_snap_cache_root()

    def _setup_snap_cache_root(self):
        snap_cache_root = os.path.join(self.project_cache_root, "snap_hashes")
        os.makedirs(snap_cache_root, exist_ok=True)
        return snap_cache_root

    def _get_snap_deb_arch(self, snap_filename):
        with tempfile.TemporaryDirectory() as temp_dir:
            unsquashfs_path = file_utils.get_tool_path("unsquashfs")
            output = subprocess.check_output(
                [
                    unsquashfs_path,
                    "-d",
                    os.path.join(temp_dir, "squashfs-root"),
                    snap_filename,
                    "-e",
                    os.path.join("meta", "snap.yaml"),
                ]
            )
            logger.debug(output)
            with open(
                os.path.join(temp_dir, "squashfs-root", "meta", "snap.yaml")
            ) as yaml_file:
                snap_yaml = yaml_utils.load(yaml_file)
        # XXX: add multiarch support later
        try:
            return snap_yaml["architectures"][0]
        except KeyError:
            return "all"

    def _get_snap_cache_path(self, snap_filename):
        snap_hash = file_utils.calculate_sha3_384(snap_filename)
        arch = self._get_snap_deb_arch(snap_filename)
        os.makedirs(os.path.join(self.snap_cache_root, arch), exist_ok=True)
        return os.path.join(self.snap_cache_root, arch, snap_hash)

    def cache(self, *, snap_filename):
        """Cache snap revision by sha3-384 hash in XDG cache, unless it already exists.
        :returns: path to cached revision.
        """
        cached_snap_path = self._get_snap_cache_path(snap_filename)
        try:
            if not os.path.isfile(cached_snap_path):
                # this must not be hard-linked, as rebuilding a snap
                # with changes should invalidate the cache, hence avoids
                # using fileutils.link_or_copy.
                shutil.copyfile(snap_filename, cached_snap_path)
        except OSError:
            logger.warning("Unable to cache snap {}.".format(snap_filename))
        return cached_snap_path

    def get(self, *, deb_arch, snap_hash=None):
        """Get the revision by sha3-384 hash or the latest cached item.

        :deb_arch: arch as string.
        :snap_hash: get by sha3 384 hash.

        :returns: full path to cached snap.
        """
        snap_cache_dir = os.path.join(self.snap_cache_root, deb_arch)
        if not os.path.isdir(snap_cache_dir):
            return None

        cached_hashes = os.listdir(snap_cache_dir)
        if not cached_hashes:
            return None

        if snap_hash:
            for cached_hash in cached_hashes:
                if cached_hash == snap_hash:
                    return os.path.join(snap_cache_dir, cached_hash)
            return None

        cached_snaps = [os.path.join(snap_cache_dir, f) for f in cached_hashes]
        return max(cached_snaps, key=os.path.getctime)

    def prune(self, *, deb_arch, keep_hash):
        """Prune the snap revisions beside the keep_hash in XDG cache.

        :returns: pruned files paths list.
        """

        pruned_files_list = []

        snap_cache_dir = os.path.join(self.snap_cache_root, deb_arch)
        for cached_hash in os.listdir(snap_cache_dir):
            if cached_hash != keep_hash:
                try:
                    cached_snap = os.path.join(snap_cache_dir, cached_hash)
                    os.remove(cached_snap)
                    pruned_files_list.append(cached_snap)
                except OSError:
                    logger.warning("Unable to prune snap {}.".format(cached_snap))
        return pruned_files_list



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.extractors
from snapcraft.internal.states._state import PartState


class PrimeState(PartState):
    yaml_tag = u"!PrimeState"

    def __init__(
        self,
        files,
        directories,
        dependency_paths=None,
        part_properties=None,
        project=None,
        scriptlet_metadata=None,
    ):
        super().__init__(part_properties, project)

        if not scriptlet_metadata:
            scriptlet_metadata = snapcraft.extractors.ExtractedMetadata()

        self.files = files
        self.directories = directories
        self.dependency_paths = set()
        self.scriptlet_metadata = scriptlet_metadata

        if dependency_paths:
            self.dependency_paths = dependency_paths

    def properties_of_interest(self, part_properties):
        """Extract the properties concerning this step from part_properties.

        The only property of interest to the prime step is the `prime` keyword
        used to filter out files with a white or blacklist.
        """

        return {
            "override-prime": part_properties.get("override-prime"),
            "prime": part_properties.get("prime", ["*"]) or ["*"],
        }

    def project_options_of_interest(self, project):
        """Extract the options concerning this step from the project.

        The prime step doesn't care about any project options.
        """

        return {}
